# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

from typing import Dict, List

from ..utils.record import MetricMeter


class BaseMetric:
    def __init__(self, metric_names, primary_metric, **kwargs):
        assert primary_metric in metric_names
        self.metric_names = metric_names
        self.primary_metric = primary_metric
        self.metric_meter = {metric_name: MetricMeter(metric_name, ':6.2f') for metric_name in metric_names}

    def calibrate(self, pred, target, **kwargs):
        pass



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

from typing import List 
from omegaconf import DictConfig
import torch 
import torch.nn as nn 
import math

from ....op.custom import SeparableConvLayer
from ....utils import AnchorBasedDetectionModelOutput 
from .detection import AnchorGenerator


class YOLOFastestHeadV2(nn.Module): 
    def __init__(
        self,
        num_classes: int, 
        intermediate_features_dim: List[int], 
        params: DictConfig) -> None:
        super().__init__()
        num_anchors = 3 # TODO
        anchors = params.anchors
        num_anchors = len(anchors[0]) // 2
        self.anchors = anchors
        tmp_cell_anchors = []
        for a in self.anchors: 
            a = torch.tensor(a).view(-1, 2)
            wa = a[:, 0:1]
            ha = a[:, 1:]
            base_anchors = torch.cat([-wa, -ha, wa, ha], dim=-1)/2
            tmp_cell_anchors.append(base_anchors) 
        self.anchor_generator = AnchorGenerator(sizes=((128),)) # TODO: dynamic image_size, and anchor_size as a parameters
        self.anchor_generator.cell_anchors = tmp_cell_anchors
        num_anchors = self.anchor_generator.num_anchors_per_location()[0]
        in_channel = intermediate_features_dim[0]
        self.cls_head = YOLOFastestClassificationHead(in_channel, num_anchors, num_classes)  
        self.reg_head = YOLOFastestRegressionHead(in_channel, num_anchors) 

    def forward(self, x): 
        anchors = torch.cat(self.anchor_generator(x), dim=0)
        cls_logits, objs = self.cls_head(x) 
        bbox_regression = self.reg_head(x)
        return AnchorBasedDetectionModelOutput(anchors=anchors, cls_logits=cls_logits, bbox_regression=bbox_regression)

def yolo_fastest_head_v2(num_classes, intermediate_features_dim, conf_model_head) -> YOLOFastestHeadV2:
    return YOLOFastestHeadV2(num_classes=num_classes, 
                             intermediate_features_dim=intermediate_features_dim,
                             params=conf_model_head.params)


class YOLOFastestClassificationHead(nn.Module):
    def __init__(
        self, 
        in_channels,
        num_anchors, 
        num_classes,
        prior_prob = 0.01, 
        ) -> None:
        super().__init__()

        self.layer_1 = nn.Sequential(*[
            SeparableConvLayer(in_channels, in_channels, 5, padding=2, no_out_act=True),
            SeparableConvLayer(in_channels, in_channels, 5, padding=2, no_out_act=True),
        ])
        self.layer_2 = nn.Sequential(*[
            SeparableConvLayer(in_channels, in_channels, 5, padding=2, no_out_act=True),
            SeparableConvLayer(in_channels, in_channels, 5, padding=2, no_out_act=True),
        ])
        
        self.cls_logits = nn.Conv2d(in_channels, num_classes * num_anchors, 1, 1, 0, bias=True)
        self.obj = nn.Conv2d(in_channels, num_anchors, 1, 1, 0, bias=True) 
        nn.init.normal_(self.cls_logits.weight, std=0.01)
        nn.init.constant_(self.cls_logits.bias, -math.log((1 - prior_prob) / prior_prob))

        self.num_classes = num_classes 
        self.num_anchors = num_anchors
    
    def forward(self, x): 
        all_cls_logits = [] 
        all_objs = []
        out1 = self.layer_1(x[0])
        out2 = self.layer_2(x[1])
        outputs = [out1, out2]

        for idx, features in enumerate(x): 
            cls_logits = outputs[idx]
            objectness = cls_logits
            cls_logits = self.cls_logits(cls_logits)
            objectness = self.obj(objectness)
            
            # Permute classification output from (N, A * K, H, W) to (N, HWA, K).
            N, _, H, W = cls_logits.shape
            cls_logits = cls_logits.view(N, -1, self.num_classes, H, W)
            cls_logits = cls_logits.permute(0, 3, 4, 1, 2)
            cls_logits = cls_logits.reshape(N, -1, self.num_classes)  # Size=(N, HWA, K)

            all_cls_logits.append(cls_logits)

            # Permute objectness output from (N, A, H, W) to (N, HWA, 1).
            N, _, H, W = objectness.shape
            objectness = objectness.view(N, -1, 1, H, W)
            objectness = objectness.permute(0, 3, 4, 1, 2)
            objectness = objectness.reshape(N, -1, 1)  # Size=(N, HWA, 1)
            all_objs.append(objectness)
        
        return all_cls_logits, all_objs



class YOLOFastestRegressionHead(nn.Module): 
    def __init__(
        self,
        in_channels, 
        num_anchors,) -> None:
        super().__init__()

        self.layer_1 = nn.Sequential(*[
            SeparableConvLayer(in_channels, in_channels, 5, padding=2, no_out_act=True),
            SeparableConvLayer(in_channels, in_channels, 5, padding=2, no_out_act=True),
        ])
        self.layer_2 = nn.Sequential(*[
            SeparableConvLayer(in_channels, in_channels, 5, padding=2, no_out_act=True),
            SeparableConvLayer(in_channels, in_channels, 5, padding=2, no_out_act=True),
        ])
        self.bbox_reg = nn.Conv2d(in_channels, num_anchors * 4, 1, 1, 0, bias=True)
    
    def forward(self, x, targets=None): 
        all_bbox_regression = []
        out1 = self.layer_1(x[0])
        out2 = self.layer_2(x[1])
        outputs = [out1, out2]

        for idx, features in enumerate(x): 
            bbox_regression = outputs[idx]
            bbox_regression = self.bbox_reg(bbox_regression)

            # Permute bbox regression output from (N, 4 * A, H, W) to (N, HWA, 4).
            N, _, H, W = bbox_regression.shape
            bbox_regression = bbox_regression.view(N, -1, 4, H, W)
            bbox_regression = bbox_regression.permute(0, 3, 4, 1, 2)
            bbox_regression = bbox_regression.reshape(N, -1, 4)  # Size=(N, HWA, 4)

            all_bbox_regression.append(bbox_regression)
        
        return all_bbox_regression




# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

from functools import partial

import cv2
import numpy as np
import PIL.Image as Image

from ..utils.constants import IMAGENET_DEFAULT_MEAN, IMAGENET_DEFAULT_STD
from .custom import image_proc as TC
from .registry import TRANSFORM_DICT

EDGE_SIZE = 4
Y_K_SIZE = 6
X_K_SIZE = 6


def reduce_label(label: np.ndarray) -> Image.Image:
    label[label == 0] = 255
    label = label - 1
    label[label == 254] = 255
    return Image.fromarray(label)


def generate_edge(label: np.ndarray) -> Image.Image:
    edge = cv2.Canny(label, 0.1, 0.2)
    kernel = np.ones((EDGE_SIZE, EDGE_SIZE), np.uint8)
    # edge_pad == True
    edge = edge[Y_K_SIZE:-Y_K_SIZE, X_K_SIZE:-X_K_SIZE]
    edge = np.pad(edge, ((Y_K_SIZE, Y_K_SIZE), (X_K_SIZE, X_K_SIZE)), mode='constant')
    edge = (cv2.dilate(edge, kernel, iterations=1) > 50) * 1.0
    return Image.fromarray((edge.copy() * 255).astype(np.uint8))


def transforms_check(transforms):
    names = [t.name.lower() for t in transforms]
    if 'mixing' in names:
        if names[-1] == 'mixing':
            return transforms[:-1]
        else:
            raise ValueError("Mixing transform is in the middle of transforms. This must be in the last of transforms list.")
    return transforms


def transforms_custom(conf_augmentation, training):
    phase_conf = conf_augmentation.train if training else conf_augmentation.inference

    preprocess = []
    if phase_conf:
        checked_transforms = transforms_check(phase_conf)
        for augment in checked_transforms:
            name = augment.name.lower()
            augment_kwargs = list(augment.keys())
            augment_kwargs.remove('name')
            augment_kwargs = {k:augment[k] for k in augment_kwargs}
            transform = TRANSFORM_DICT[name](**augment_kwargs)
            preprocess.append(transform)

    return TC.Compose(preprocess)


def train_transforms_pidnet(conf_augmentation, training):
    preprocess = []
    phase_conf = conf_augmentation.train if training else conf_augmentation.inference

    if phase_conf:
        checked_transforms = transforms_check(phase_conf)
        for augment in checked_transforms:
            name = augment.name.lower()
            augment_kwargs = list(augment.keys())
            augment_kwargs.remove('name')
            augment_kwargs = {k:augment[k] for k in augment_kwargs}
            transform = TRANSFORM_DICT[name](**augment_kwargs)
            preprocess.append(transform)

    return TC.Compose(preprocess, additional_targets={'edge': 'mask'})


def create_transform(model_name: str, is_training=False):
    if 'pidnet' in model_name:
        return partial(train_transforms_pidnet, training=is_training)
    return partial(transforms_custom, training=is_training)



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

import os
from abc import ABC, abstractmethod, abstractproperty
from itertools import repeat
from pathlib import Path

import numpy as np
import torch
import torch.utils.data as data
from loguru import logger


class BaseCustomDataset(data.Dataset):

    def __init__(self, conf_data, conf_augmentation, model_name, idx_to_class, split, samples, transform, **kwargs):
        super(BaseCustomDataset, self).__init__()
        self.conf_data = conf_data
        self.conf_augmentation = conf_augmentation
        self.model_name = model_name

        self.transform = transform(conf_augmentation)
        self.samples = samples

        self._root = conf_data.path.root
        self._idx_to_class = idx_to_class
        self._num_classes = len(self._idx_to_class)
        self._split = split

        self.cache = False

    @abstractmethod
    def __getitem__(self, index):
        pass

    @abstractmethod
    def cache_dataset(self, sampler, distributed):
        pass

    def __len__(self):
        return len(self.samples)

    @property
    def num_classes(self):
        return self._num_classes

    @property
    def class_map(self):
        return self._idx_to_class

    @property
    def root(self):
        return self._root

    @property
    def mode(self):
        return self._split


class BaseHFDataset(data.Dataset):

    def __init__(self, conf_data, conf_augmentation, model_name, root, split, transform):
        super(BaseHFDataset, self).__init__()
        self.conf_data = conf_data
        self.conf_augmentation = conf_augmentation
        self.model_name = model_name
        self.transform = transform(conf_augmentation)
        self._root = root
        self._split = split

    def _load_dataset(self, root, subset_name=None, cache_dir=None):
        from datasets import load_dataset
        if cache_dir is not None:
            Path(cache_dir).mkdir(exist_ok=True, parents=True)
        total_dataset = load_dataset(root, name=subset_name, cache_dir=cache_dir)
        return total_dataset

    @abstractmethod
    def __getitem__(self, index):
        pass

    @abstractmethod
    def __len__(self):
        pass

    @abstractproperty
    def num_classes(self):
        pass

    @abstractproperty
    def class_map(self):
        pass

    @property
    def root(self):
        return self._name

    @property
    def mode(self):
        return self._split


class BaseSampleLoader(ABC):
    def __init__(self, conf_data, train_valid_split_ratio):
        self.conf_data = conf_data
        self.train_valid_split_ratio = train_valid_split_ratio

    @abstractmethod
    def load_data(self):
        raise NotImplementedError

    @abstractmethod
    def load_id_mapping(self):
        raise NotImplementedError

    @abstractmethod
    def load_class_map(self, id_mapping):
        raise NotImplementedError

    def load_samples(self):
        assert self.conf_data.id_mapping is not None
        id_mapping = self.load_id_mapping()
        misc = self.load_class_map(id_mapping)

        train_samples, valid_samples, test_samples = self.load_split_samples()
        return train_samples, valid_samples, test_samples, misc

    def load_split_samples(self):
        exists_train = self.conf_data.path.train.image is not None
        exists_valid = self.conf_data.path.valid.image is not None
        exists_test = self.conf_data.path.test.image is not None

        train_samples = None
        valid_samples = None
        test_samples = None

        if exists_train:
            train_samples = self.load_data(split='train')
        if exists_valid:
            valid_samples = self.load_data(split='valid')
        if exists_test:
            test_samples = self.load_data(split='test')

        if not exists_valid and exists_train:
            logger.info(f"Validation set is not provided in config. Split automatically training set by {self.train_valid_split_ratio:.1f}:{1-self.train_valid_split_ratio:.1f}.")
            num_train_splitted = int(len(train_samples) * self.train_valid_split_ratio)
            train_samples, valid_samples = \
                data.random_split(train_samples, [num_train_splitted, len(train_samples) - num_train_splitted],
                                  generator=torch.Generator().manual_seed(42))

        return train_samples, valid_samples, test_samples

    @abstractmethod
    def load_huggingface_samples(self):
        raise NotImplementedError



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

import math
import warnings

import torch
from loguru import logger
from torch.optim.lr_scheduler import _LRScheduler


class CosineAnnealingLRWithCustomWarmUp(_LRScheduler):
    """Modified from CosineAnnealingLR in PyTorch

    Args:
        optimizer (Optimizer): Wrapped optimizer.
        warmup_iters (int): The number of steps that the scheduler finishes to warmup the learning rate.
        total_iters (int): Maximum number of iterations. Originally named as `T_max`.
        warmup_bias_lr (int): Starting learning rate for warmup period.
        min_lr (float): Minimum learning rate. Originally named as `eta_min`.
    """

    def __init__(
        self,
        optimizer,
        scheduler_conf,
        training_epochs,
    ):
        if scheduler_conf.end_epoch > training_epochs:
            logger.warning('``training.scheduler.end_epoch`` is larger than ``training.epochs``. Automatically set scheduler end_epoch as training epochs.')
            scheduler_conf.end_epoch = training_epochs
        self.T_max = scheduler_conf.end_epoch
        self.eta_min = scheduler_conf.min_lr
        self.warmup_bias_lr = scheduler_conf.warmup_bias_lr
        self.warmup_iters = scheduler_conf.warmup_epochs
        super().__init__(optimizer)

    def get_lr(self):
        if not self._get_lr_called_within_step:
            warnings.warn("To get the last learning rate computed by the scheduler, "
                          "please use `get_last_lr()`.", UserWarning, stacklevel=2)

        if self.last_epoch > self.T_max:
            return [group['lr'] for group in self.optimizer.param_groups]

        if self.last_epoch >= 0 and self.last_epoch < self.warmup_iters:
            return [self.warmup_bias_lr + (float(self.last_epoch + 1) / float(max(1, self.warmup_iters))) * (base_lr - self.warmup_bias_lr)
                    for base_lr in self.base_lrs]

        if self._step_count == 1 and self.last_epoch > 0:
            return [self.eta_min + (base_lr - self.eta_min) *
                    (1 + math.cos((self.last_epoch) * math.pi / self.T_max)) / 2
                    for base_lr, group in
                    zip(self.base_lrs, self.optimizer.param_groups)]
        elif (self.last_epoch - 1 - self.T_max) % (2 * self.T_max) == 0:
            return [group['lr'] + (base_lr - self.eta_min) *
                    (1 - math.cos(math.pi / self.T_max)) / 2
                    for base_lr, group in
                    zip(self.base_lrs, self.optimizer.param_groups)]
        return [(1 + math.cos(math.pi * self.last_epoch / self.T_max)) /
                (1 + math.cos(math.pi * (self.last_epoch - 1) / self.T_max)) *
                (group['lr'] - self.eta_min) + self.eta_min
                for group in self.optimizer.param_groups]

    def _get_closed_form_lr(self):
        return [
            (
                min(
                    self.warmup_bias_lr + (float(self.last_epoch + 1) / float(max(1, self.warmup_iters))) * (base_lr - self.warmup_bias_lr),
                    self.eta_min + (base_lr - self.eta_min) * (1 + math.cos(math.pi * self.last_epoch / self.T_max)) / 2
                )
            )
            for base_lr in self.base_lrs
        ]



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

from pathlib import Path
from typing import Any, List, Optional, TypedDict, Union

import omegaconf
import torch
import torch.nn as nn
from loguru import logger
from torch import Tensor
from torch.fx.proxy import Proxy

from ..utils.checkpoint import load_checkpoint

FXTensorType = Union[Tensor, Proxy]
FXTensorListType = Union[List[Tensor], List[Proxy]]

DEFAULT_CACHE_DIR = Path.home() / ".cache" / f"{__name__.split('.')[0]}"  # ~/.cache/netspresso_trainer

DEFAULT_WEIGHT_VERSION_DICT = {
    'resnet18': 'imagenet1k',
    'resnet34': 'imagenet1k',
    'resnet50': 'imagenet1k',
    'mobilenet_v3_small': 'imagenet1k',
    'mobilenet_v3_large': 'imagenet1k',
    'mobilenet_v4_conv_small': 'imagenet1k',
    'mobilenet_v4_conv_medium': 'imagenet1k',
    'mobilenet_v4_conv_large': 'imagenet1k',
    'mobilenet_v4_hybrid_medium': 'imagenet1k',
    'mobilenet_v4_hybrid_large': 'imagenet1k',
    'segformer_b0': 'ade20k',
    'mobilevit_s': 'imagenet1k',
    'vit_tiny': 'imagenet1k',
    'efficientformer_l1': 'imagenet1k',
    'mixnet_s': 'imagenet1k',
    'mixnet_m': 'imagenet1k',
    'mixnet_l': 'imagenet1k',
    'pidnet_s': 'cityscapes',
    'yolox_nano': 'coco',
    'yolox_tiny': 'coco',
    'yolox_s': 'coco',
    'yolox_m': 'coco',
    'yolox_l': 'coco',
    'yolox_x': 'coco',
    'rtdetr_res18': 'coco',
    'rtdetr_res50': 'coco',
}

MODEL_CHECKPOINT_URL_DICT = {
    'resnet18': {
        'imagenet1k': "https://netspresso-trainer-public.s3.ap-northeast-2.amazonaws.com/checkpoint/resnet/resnet18_imagenet1k.safetensors?versionId=mEn38lTgeWB_kkDQOdiF9DfQKrPL2CCy",
    },
    'resnet34': {
        'imagenet1k': "https://netspresso-trainer-public.s3.ap-northeast-2.amazonaws.com/checkpoint/resnet/resnet34_imagenet1k.safetensors?versionId=.6Ezbpm8lsRyW.HrrnL6AyywYnsjlarr",
    },
    'resnet50': {
        'imagenet1k': "https://netspresso-trainer-public.s3.ap-northeast-2.amazonaws.com/checkpoint/resnet/resnet50_imagenet1k.safetensors?versionId=UF509XgguL1T3IhxSw1HVpfvnhvGGy6J",
    },
    'mobilenet_v3_small': {
        'imagenet1k': "https://netspresso-trainer-public.s3.ap-northeast-2.amazonaws.com/checkpoint/mobilenetv3/mobilenet_v3_small_imagenet1k.safetensors?versionId=dPgl9WzLaJLqmuyDxsCoYawItdi0zES8",
    },
    'mobilenet_v3_large': {
        'imagenet1k': "https://netspresso-trainer-public.s3.ap-northeast-2.amazonaws.com/checkpoint/mobilenetv3/mobilenet_v3_large_imagenet1k.safetensors?versionId=jPG4LAueBDO5VrFGLQ51_z.iDHa5lOgP",
    },
    'mobilenet_v4_conv_small': {
        'imagenet1k': "https://netspresso-trainer-public.s3.ap-northeast-2.amazonaws.com/checkpoint/mobilenetv4/mobilenet_v4_conv_small_imagenet1k.safetensors?versionId=0bpPNyhCNfF5FzHSnXkJFgP8pyU34GKt",
    },
    'mobilenet_v4_conv_medium': {
        'imagenet1k': "https://netspresso-trainer-public.s3.ap-northeast-2.amazonaws.com/checkpoint/mobilenetv4/mobilenet_v4_conv_medium_imagenet1k.safetensors?versionId=buTzldKEk8MSWZHehi494KsNMfP3G1Zr",
    },
    'mobilenet_v4_conv_large': {
        'imagenet1k': "https://netspresso-trainer-public.s3.ap-northeast-2.amazonaws.com/checkpoint/mobilenetv4/mobilenet_v4_conv_large_imagenet1k.safetensors?versionId=_5D7G_yhUg2YJqwBBgNgUxCHSQPbERSD",
    },
    'mobilenet_v4_hybrid_medium': {
        'imagenet1k': "https://netspresso-trainer-public.s3.ap-northeast-2.amazonaws.com/checkpoint/mobilenetv4/mobilenet_v4_hybrid_medium_imagenet1k.safetensors?versionId=5eBSYAwF.HjVeTWTOS.YFeL3f_FIa6Nv",
    },
    'mobilenet_v4_hybrid_large': {
        'imagenet1k': "https://netspresso-trainer-public.s3.ap-northeast-2.amazonaws.com/checkpoint/mobilenetv4/mobilenet_v4_hybrid_large_imagenet1k.safetensors?versionId=UAbo2Ag4dMiOPdgO3qYl2ztK4Vn9PnAv",
    },
    'segformer_b0': {
        'ade20k': "https://netspresso-trainer-public.s3.ap-northeast-2.amazonaws.com/checkpoint/segformer/segformer_b0_ade20k.safetensors?versionId=0RRDpZeHb2VvVzFo2jGZN4A4bVQ.k49l",
    },
    'mobilevit_s': {
        'imagenet1k': "https://netspresso-trainer-public.s3.ap-northeast-2.amazonaws.com/checkpoint/mobilevit/mobilevit_s_imagenet1k.safetensors?versionId=IvxVWQ.yqTF9tvZr9E2JLyE7_1dBdDB4",
    },
    'vit_tiny': {
        'imagenet1k': "https://netspresso-trainer-public.s3.ap-northeast-2.amazonaws.com/checkpoint/vit/vit_tiny_imagenet1k.safetensors?versionId=FJwTnbWvFxnlIK.57wjWCh517kuFpOkF",
    },
    'efficientformer_l1': {
        'imagenet1k': "https://netspresso-trainer-public.s3.ap-northeast-2.amazonaws.com/checkpoint/efficientformer/efficientformer_l1_imagenet1k.safetensors?versionId=U5gtiRXNNyBvXOpawPA7lfFil09rq50S",
    },
    'mixnet_s': {
        'imagenet1k': "https://netspresso-trainer-public.s3.ap-northeast-2.amazonaws.com/checkpoint/mixnet/mixnet_s_imagenet1k.safetensors?versionId=9YlFMjHa__1GYFv1H4l4_d4NqX_5djT5",
    },
    'mixnet_m': {
        'imagenet1k': "https://netspresso-trainer-public.s3.ap-northeast-2.amazonaws.com/checkpoint/mixnet/mixnet_m_imagenet1k.safetensors?versionId=F11svT5UpQJWGQTv9B..STYj5_P53O1Y",
    },
    'mixnet_l': {
        'imagenet1k': "https://netspresso-trainer-public.s3.ap-northeast-2.amazonaws.com/checkpoint/mixnet/mixnet_l_imagenet1k.safetensors?versionId=nLUZeSGKWRZVldPLc0FCvFCZc6lPPEyB",
    },
    'pidnet_s': {
        'cityscapes': "https://netspresso-trainer-public.s3.ap-northeast-2.amazonaws.com/checkpoint/pidnet/pidnet_s_cityscapes.safetensors?versionId=lsgtDpiF1yqJpuCLYpruLdR6on0V53r8",
    },
    'yolox_nano': {
        'coco': "https://netspresso-trainer-public.s3.ap-northeast-2.amazonaws.com/checkpoint/yolox/yolox_nano_coco.safetensors?versionId=JCXugDTwGegx9Kl6Jc5AMJpIkA.WlNVP"
    },
    'yolox_tiny': {
        'coco': "https://netspresso-trainer-public.s3.ap-northeast-2.amazonaws.com/checkpoint/yolox/yolox_tiny_coco.safetensors?versionId=lJp1bCEToD_6IaL9kRCqcYIwVZ.QQ.1P"
    },
    'yolox_s': {
        'coco': "https://netspresso-trainer-public.s3.ap-northeast-2.amazonaws.com/checkpoint/yolox/yolox_s_coco.safetensors?versionId=QRLqHKqhv8TSYBrmsQ3M8lCR8w7HEZyA",
    },
    'yolox_m': {
        'coco': "https://netspresso-trainer-public.s3.ap-northeast-2.amazonaws.com/checkpoint/yolox/yolox_m_coco.safetensors?versionId=xVUySP8xgVTpa6NhCMQpulqmYeRUAhpS",
    },
    'yolox_l': {
        'coco': "https://netspresso-trainer-public.s3.ap-northeast-2.amazonaws.com/checkpoint/yolox/yolox_l_coco.safetensors?versionId=1GR6YNRu.yUfnjq8hKPgARyZ6YejdxMB",
    },
    'yolox_x': {
        'coco': "https://netspresso-trainer-public.s3.ap-northeast-2.amazonaws.com/checkpoint/yolox/yolox_x_coco.safetensors?versionId=NWskUEbSGviBWskHQ3P1dQZXnRXOR1WN",
    },
    'rtdetr_res18': {
        'coco': "https://netspresso-trainer-public.s3.ap-northeast-2.amazonaws.com/checkpoint/rtdetr/rtdetr_res18_coco.safetensors?versionId=uu9v49NI6rQx8wOY6bJbEXUFOG_R9xqH",
    },
    'rtdetr_res50': {
        'coco': "https://netspresso-trainer-public.s3.ap-northeast-2.amazonaws.com/checkpoint/rtdetr/rtdetr_res50_coco.safetensors?versionId=JHmnjY13BEflpnDCYPFJ1c17UwpqDrLQ",
    },
}


class BackboneOutput(TypedDict):
    intermediate_features: Optional[FXTensorListType]
    last_feature: Optional[FXTensorType]


class ModelOutput(TypedDict):
    pred: FXTensorType


class AnchorBasedDetectionModelOutput(ModelOutput):
    anchors: FXTensorType
    cls_logits: FXTensorType
    bbox_regression: FXTensorType


class DetectionModelOutput(ModelOutput):
    boxes: Any
    proposals: Any
    anchors: Any
    objectness: Any
    pred_bbox_detlas: Any
    class_logits: Any
    box_regression: Any
    labels: Any
    regression_targets: Any
    post_boxes: Any
    post_scores: Any
    post_labels: Any


class PIDNetModelOutput(ModelOutput):
    extra_p: Optional[FXTensorType]
    extra_d: Optional[FXTensorType]


def download_model_checkpoint(
    model_name: str,
    task: Optional[str] = None,  # TODO: Pretrained weights can be distinguished by task
) -> Path:
    assert model_name in DEFAULT_WEIGHT_VERSION_DICT
    assert model_name in MODEL_CHECKPOINT_URL_DICT

    # TODO: User can select the specific weight version
    checkpoint_weight_version = DEFAULT_WEIGHT_VERSION_DICT[model_name]

    checkpoint_url = MODEL_CHECKPOINT_URL_DICT[model_name][checkpoint_weight_version]

    checkpoint_filename = Path(checkpoint_url).name.split('?versionId')[0] # @illian01: Remove specified version id
    model_checkpoint: Path = DEFAULT_CACHE_DIR / model_name / checkpoint_filename
    model_checkpoint.parent.mkdir(parents=True, exist_ok=True)
    # Safer switch: only extension, user can use the custom name for checkpoint file
    model_checkpoint = model_checkpoint.with_suffix(Path(checkpoint_filename).suffix)
    if not model_checkpoint.exists():
        torch.hub.download_url_to_file(checkpoint_url, model_checkpoint)

    return model_checkpoint


def load_from_checkpoint(
    model: nn.Module,
    model_checkpoint: Optional[Union[str, Path]],
    load_checkpoint_head: bool,
) -> nn.Module:
    model_name = model.name
    task = model.task

    if model_checkpoint is None:
        assert model_name is not None, "When `use_pretrain` is True, model_name should be given."
        assert model_name in MODEL_CHECKPOINT_URL_DICT, \
            f"model_name {model_name} in path {model_checkpoint} is not valid name!"
        model_checkpoint = download_model_checkpoint(model_name, task)
        logger.info(f"Pretrained model for {model_name} is loaded from: {model_checkpoint}")

    model_state_dict = load_checkpoint(model_checkpoint)
    if not load_checkpoint_head:
        logger.info("-"*40)
        logger.info("Head weights are not loaded because model.checkpoint.load_head is set to False")
        head_keys = [key for key in model_state_dict if key.startswith(model.head_list)]
        for key in head_keys:
            del model_state_dict[key]

    missing_keys, unexpected_keys = model.load_state_dict(model_state_dict, strict=False)

    if not load_checkpoint_head:
        missing_keys = [key for key in missing_keys if not key.startswith(model.head_list)]

    if len(missing_keys) != 0:
        logger.warning(f"Missing key(s) in state_dict: {missing_keys}")
    if len(unexpected_keys) != 0:
        logger.warning(f"Unexpected key(s) in state_dict: {unexpected_keys}")

    return model


def is_single_task_model(model_conf: omegaconf.DictConfig):
    conf_model_architecture_full = model_conf.architecture.full
    if conf_model_architecture_full is None:
        return False
    if conf_model_architecture_full.name is None:
        return False
    return True

def get_model_format(model_conf: omegaconf.DictConfig):
    if not model_conf.checkpoint.path:
        return 'torch'

    model_path = Path(model_conf.checkpoint.path)
    ext = model_path.suffix

    if ext == '.safetensors':
        return 'torch'
    elif ext == '.pt':
        return 'torch.fx'
    elif ext == '.onnx':
        return 'onnx'
    else:
        raise ValueError(f"Unsupported model format: {model_conf.checkpoint.path}")



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

from typing import Callable, Dict, Literal, Type

from .base import BaseMetric
from .classification import ClassificationMetric
from .detection import DetectionMetric
from .pose_estimation import PoseEstimationMetric
from .segmentation import SegmentationMetric

TASK_METRIC: Dict[Literal['classification', 'segmentation', 'detection'], Type[BaseMetric]] = {
    'classification': ClassificationMetric,
    'segmentation': SegmentationMetric,
    'detection': DetectionMetric,
    'pose_estimation': PoseEstimationMetric,
}

PHASE_LIST = ['train', 'valid', 'test']



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

from pathlib import Path
from typing import Any, Dict, List, Literal, Optional, Tuple, Union

import numpy as np
import PIL.Image as Image


class ImageSaver:
    def __init__(self, model, result_dir) -> None:
        super(ImageSaver, self).__init__()
        self.model = model
        self.save_dir: Path = Path(result_dir) / "result_image"
        self.save_dir.mkdir(exist_ok=True)

    def save_ndarray_as_image(self, image_array: np.ndarray, filename: Union[str, Path], dataformats: Literal['HWC', 'CHW'] = 'HWC'):
        assert image_array.ndim == 3
        if dataformats != 'HWC' and dataformats == 'CHW':
            image_array = image_array.transpose((1, 2, 0))

        # HWC
        assert image_array.shape[-1] in [1, 3]
        Image.fromarray(image_array.astype(np.uint8)).save(filename)
        return True

    def save_result(self, image_dict: Dict, prefix, epoch):
        prefix_dir: Path = self.save_dir / prefix
        prefix_dir.mkdir(exist_ok=True)

        for k, v_list in image_dict.items():
            for idx, v in enumerate(v_list):
                assert isinstance(v, np.ndarray)
                if epoch is None:
                    self.save_ndarray_as_image(v, f"{prefix_dir}/{idx:03d}_{k}.png", dataformats='HWC')
                else:
                    self.save_ndarray_as_image(v, f"{prefix_dir}/{epoch:04d}_{idx:03d}_{k}.png", dataformats='HWC')

    def __call__(
        self,
        prefix: Literal['training', 'validation', 'evaluation', 'inference'],
        epoch: Optional[int] = None,
        images: Optional[List] = None,
        **kwargs
    ):
        if images is not None:
            self.save_result(images, prefix=prefix, epoch=epoch)



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

from pathlib import Path
from typing import Callable, Dict, List, Type

import torch.nn as nn

from .backbones import (
    cspdarknet,
    efficientformer,
    mixnet,
    mixtransformer,
    mobilenetv3,
    mobilenetv4,
    mobilevit,
    resnet,
    shufflenetv2,
    vit,
)
from .base import ClassificationModel, DetectionModel, PoseEstimationModel, SegmentationModel, TaskModel
from .full import pidnet
from .heads.classification import fc, fc_conv
from .heads.detection import anchor_decoupled_head, anchor_free_decoupled_head, rtdetr_head, yolo_fastest_head_v2
from .heads.pose_estimation import rtmcc
from .heads.segmentation import all_mlp_decoder
from .necks import fpn, lightfpn, rtdetr_hybrid_encoder, yolopafpn

MODEL_BACKBONE_DICT: Dict[str, Callable[..., nn.Module]] = {
    'resnet': resnet,
    'mobilenetv3': mobilenetv3,
    'mobilenetv4': mobilenetv4,
    'mixtransformer': mixtransformer,
    'mobilevit': mobilevit,
    'vit': vit,
    'efficientformer': efficientformer,
    'cspdarknet': cspdarknet,
    'shufflenetv2': shufflenetv2,
    'mixnet': mixnet,
}

MODEL_NECK_DICT: Dict[str, Callable[..., nn.Module]] = {
    'fpn': fpn,
    'lightfpn': lightfpn,
    'yolopafpn': yolopafpn,
    'rtdetr_hybrid_encoder': rtdetr_hybrid_encoder,
}

MODEL_HEAD_DICT: Dict[str, Callable[..., nn.Module]] = {
    'classification': {
        'fc': fc,
        'fc_conv': fc_conv,
    },
    'segmentation': {
        'all_mlp_decoder': all_mlp_decoder,
    },
    'detection': {
        'anchor_free_decoupled_head': anchor_free_decoupled_head,
        'anchor_decoupled_head': anchor_decoupled_head,
        'yolo_fastest_head_v2': yolo_fastest_head_v2,
        'rtdetr_head': rtdetr_head
    },
    'pose_estimation': {
        'rtmcc': rtmcc,
    }
}

MODEL_FULL_DICT = {
    'pidnet': pidnet
}

SUPPORTING_MODEL_LIST = list(MODEL_BACKBONE_DICT.keys()) + list(MODEL_FULL_DICT.keys())
SUPPORTING_TASK_LIST: List[str] = ['classification', 'segmentation', 'detection', 'pose_estimation']

TASK_MODEL_DICT: Dict[str, Type[TaskModel]] = {
    'classification': ClassificationModel,
    'segmentation': SegmentationModel,
    'detection': DetectionModel,
    'pose_estimation': PoseEstimationModel,
}



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

from typing import Dict, Optional, List 
from omegaconf import DictConfig
import torch
import torch.nn as nn 

from ...op.custom import ConvLayer, ShuffleV2Block
from ...utils import BackboneOutput
from ..registry import USE_INTERMEDIATE_FEATURES_TASK_LIST


__all__ = []
SUPPORTING_TASK = ['detection']

class ShuffleNetV2(nn.Module):
    def __init__(
        self,
        task: str,
        params: Optional[DictConfig] = None,
        stage_params: Optional[List] = None,
    ) -> None:
        # Check task compatibility 
        self.task = task.lower() 
        assert self.task in SUPPORTING_TASK, f"ShuffleNetV2 is not supported on {self.task} task now."
        self.use_intermediate_features = self.task in USE_INTERMEDIATE_FEATURES_TASK_LIST
        super().__init__()

        self.stage_repeats = [4, 8, 4]
        self.stage_out_channels = self._get_stage_out_channels(params.model_size)
        self._feature_dim = self.stage_out_channels[-1]
        self._intermediate_features_dim = self.stage_out_channels[-2:]
        
        self._build_network()
 
    def forward(self, x):
        x = self.maxpool(self.conv1(x))
        c1 = self.stage2(x)
        c2 = self.stage3(c1)
        c3 = self.stage4(c2)
        
        if self.use_intermediate_features:
            return BackboneOutput(intermediate_features=[c2, c3])
        
        x = self.avgpool(c3)
        x = torch.flatten(x, 1)
        return BackboneOutput(last_feature=x)

    def _get_stage_out_channels(self, model_size: str) -> List[int]:
        channels = {
            '0.5x': [-1, 24, 48, 96, 192],
            '1.0x': [-1, 24, 116, 232, 464],
            '1.5x': [-1, 24, 176, 352, 704],
            '2.0x': [-1, 24, 244, 488, 976]
        }
        if model_size not in channels:
            raise ValueError(f"Unsupported model_size: {model_size}. Available options are: {list(channels.keys())}")
        return channels[model_size]

    def _make_stage(self, in_channels: int, out_channels: int, num_layers: int) -> nn.Sequential:
        layers = []
        for i in range(num_layers):
            stride = 2 if i == 0 else 1
            in_ch = in_channels if i == 0 else out_channels // 2
            layers.append(ShuffleV2Block(in_ch, out_channels, 
                                         hidden_channels=out_channels // 2, 
                                         kernel_size=3, stride=stride))
        return nn.Sequential(*layers)
    
    def _build_network(self):
        in_channels = self.stage_out_channels[1]
        self.conv1 = ConvLayer(3, in_channels, kernel_size=3, stride=2, padding=1)
        self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)
        
        for i, num_layers in enumerate(self.stage_repeats):
            stage_name = f"stage{i+2}"
            out_channels = self.stage_out_channels[i+2]
            setattr(self, stage_name, self._make_stage(in_channels, out_channels, num_layers))
            in_channels = out_channels
        
        self.avgpool = nn.AdaptiveAvgPool2d(1) if not self.use_intermediate_features else None

    @property
    def feature_dim(self):
        return self._feature_dim

    @property
    def intermediate_features_dim(self):
        return self._intermediate_features_dim

    def task_support(self, task):
        return task.lower() in SUPPORTING_TASK


def shufflenetv2(task, conf_model_backbone) -> ShuffleNetV2: 
    return ShuffleNetV2(task, conf_model_backbone.params, conf_model_backbone.stage_params)




# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

from typing import Literal, Optional

import numpy as np
import torch

from .base import BaseTaskProcessor


class PoseEstimationProcessor(BaseTaskProcessor):
    def __init__(self, conf, postprocessor, devices, **kwargs):
        super(PoseEstimationProcessor, self).__init__(conf, postprocessor, devices, **kwargs)

    def train_step(self, train_model, batch, optimizer, loss_factory, metric_factory):
        train_model.train()
        images, keypoints = batch['pixel_values'], batch['keypoints']
        images = images.to(self.devices)
        target = {'keypoints': keypoints.to(self.devices)}

        optimizer.zero_grad()

        with torch.cuda.amp.autocast(enabled=self.mixed_precision):
            out = train_model(images, target=target)
            loss_factory.calc(out, target, phase='train')

        loss_factory.backward(self.grad_scaler)
        if self.max_norm:
            # Unscales the gradients of optimizer's assigned parameters in-place
            self.grad_scaler.unscale_(optimizer)
            torch.nn.utils.clip_grad_norm_(train_model.parameters(), self.max_norm)
        # optimizer's gradients are already unscaled, so scaler.step doesn't unscale them,
        self.grad_scaler.step(optimizer)
        self.grad_scaler.update()

        pred = self.postprocessor(out)

        keypoints = keypoints.detach().cpu().numpy()
        if self.conf.distributed:
            gathered_pred = [None for _ in range(torch.distributed.get_world_size())]
            gathered_labels = [None for _ in range(torch.distributed.get_world_size())]

            torch.distributed.gather_object(pred, gathered_pred if torch.distributed.get_rank() == 0 else None, dst=0)
            torch.distributed.gather_object(keypoints, gathered_labels if torch.distributed.get_rank() == 0 else None, dst=0)
            torch.distributed.barrier()
            if torch.distributed.get_rank() == 0:
                pred = np.concatenate(gathered_pred, axis=0)
                keypoints = np.concatenate(gathered_labels, axis=0)

        if self.single_gpu_or_rank_zero:
            logs = {
                'target': keypoints,
                'pred': pred
            }
            return dict(logs.items())

    def valid_step(self, eval_model, batch, loss_factory, metric_factory):
        eval_model.eval()
        indices, images, keypoints = batch['indices'], batch['pixel_values'], batch['keypoints']
        images = images.to(self.devices)
        target = {'keypoints': keypoints.to(self.devices)}

        out = eval_model(images)
        loss_factory.calc(out, target, phase='valid')

        pred = self.postprocessor(out)

        indices = indices.numpy()
        keypoints = keypoints.detach().cpu().numpy()
        if self.conf.distributed:
            pred = pred[indices != -1]
            keypoints = keypoints[indices != -1]

            gathered_pred = [None for _ in range(torch.distributed.get_world_size())]
            gathered_labels = [None for _ in range(torch.distributed.get_world_size())]

            torch.distributed.gather_object(pred, gathered_pred if torch.distributed.get_rank() == 0 else None, dst=0)
            torch.distributed.gather_object(keypoints, gathered_labels if torch.distributed.get_rank() == 0 else None, dst=0)
            torch.distributed.barrier()
            if torch.distributed.get_rank() == 0:
                pred = np.concatenate(gathered_pred, axis=0)
                keypoints = np.concatenate(gathered_labels, axis=0)

        if self.single_gpu_or_rank_zero:
            logs = {
                'images': images.detach().cpu().numpy(),
                'target': keypoints,
                'pred': pred
            }
            return dict(logs.items())

    def test_step(self, test_model, batch):
        test_model.eval()
        indices, images = batch['indices'], batch['pixel_values']
        images = images.to(self.devices)

        out = test_model(images)

        pred = self.postprocessor(out)

        indices = indices.numpy()
        if self.conf.distributed:
            pred = pred[indices != -1]

            gathered_pred = [None for _ in range(torch.distributed.get_world_size())]

            torch.distributed.gather_object(pred, gathered_pred if torch.distributed.get_rank() == 0 else None, dst=0)
            torch.distributed.barrier()
            if torch.distributed.get_rank() == 0:
                gathered_pred = sum(gathered_pred, [])
                pred = gathered_pred

        if self.single_gpu_or_rank_zero:
            results = {'images': images.detach().cpu().numpy(), 'pred': pred}
            return results

    def get_metric_with_all_outputs(self, outputs, phase: Literal['train', 'valid'], metric_factory):
        if self.single_gpu_or_rank_zero:
            pred = np.concatenate([output['pred']for output in outputs], axis=0)
            keypoints = np.concatenate([output['target']for output in outputs], axis=0)
            metric_factory.update(pred, keypoints, phase=phase)



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

import time
from dataclasses import dataclass, field
from typing import Dict, List, Optional, Union

__all__ = ['AverageMeter', 'Timer', 'TrainingSummary']


class AverageMeter(object):
    """Computes and stores the average and current value"""

    def __init__(self, name: str, fmt=':f'):
        self.name = name
        self.fmt = fmt
        self.reset()

    def reset(self):
        self._val: float = 0.
        self._avg: float = 0.
        self._sum: float = 0.
        self._count: int = 0

    def update(self, val: Union[float, int], n: int = 1) -> None:
        self._val = val
        self._sum += val * n
        self._count += n
        self._avg = self._sum / self._count

    def __str__(self):
        fmtstr = '{name} {val' + self.fmt + '} ({avg' + self.fmt + '})'
        return fmtstr.format(**self.__dict__)

    @property
    def avg(self) -> float:
        return self._avg


class MetricMeter(object):
    """Computes and stores the average and current value"""

    def __init__(self, name: str, fmt=':f'):
        self.name = name
        self.fmt = fmt
        self.reset()

    def reset(self):
        self._val: float = 0.
        self._avg: float = 0.
        self._sum: float = 0.
        self._count: int = 0

    def update(self, val: Union[float, int], n: int = 1) -> None:
        self._val = val
        self._sum += val
        self._count += n
        self._avg = self._sum / self._count

    def __str__(self):
        fmtstr = '{name} {val' + self.fmt + '} ({avg' + self.fmt + '})'
        return fmtstr.format(**self.__dict__)

    @property
    def avg(self) -> float:
        return self._avg


class TimeRecode:
    def __init__(self) -> None:
        self._start = time.time()
        self._end = None

    def end(self) -> None:
        self._end = time.time()

    @property
    def done(self) -> bool:
        return self._end is not None

    @property
    def elapsed(self) -> float:
        return self._end - self._start if self._end is not None else None


class Timer:
    """Basic timer with CRUD functions
    """

    def __init__(self) -> None:
        self.history = {}

    def start_record(self, name) -> bool:
        if name not in self.history:
            self.history[name] = TimeRecode()  # create
            return True
        return False  # fail cause alreay exists

    def _end_record(self, name) -> bool:
        if name in self.history:
            self.history[name].end()  # update
            return True  # success
        return False  # fail cause no such key

    def end_record(self, name):
        self._end_record(name)

    def get(self, name, as_pop=True) -> Optional[float]:
        if name in self.history:
            record = self.history.pop(name) if as_pop else self.history[name]  # read (+ delete)
            if not record.done:
                record.end()
                assert record.done
            return record.elapsed

        return  # no such key


TYPE_SUMMARY_RECORD = Dict[int, Union[float, Dict[str, float]]]  # {epoch: value, ...}


@dataclass
class TrainingSummary:
    total_epoch: int
    train_losses: TYPE_SUMMARY_RECORD
    valid_losses: TYPE_SUMMARY_RECORD
    train_metrics: TYPE_SUMMARY_RECORD
    valid_metrics: TYPE_SUMMARY_RECORD
    metrics_list: List[str]
    primary_metric: str
    flops: Optional[int] = None
    params: Optional[int] = None
    total_train_time: Optional[float] = None
    best_epoch: int = field(init=False)
    last_epoch: int = field(init=False)
    status: str = ""
    error_stats: str = ""

    def __post_init__(self):
        self.last_epoch = 1 if not self.train_losses else list(self.train_losses.keys())[-1] # self.train_losses is empty if error occurs before first epoch done
        self.best_epoch = self.last_epoch if not self.valid_losses else min(self.valid_losses, key=self.valid_losses.get) # self.valid_losses is empty if validation is not performed

@dataclass
class EvaluationSummary:
    losses: float
    metrics: float
    metrics_list: List[str]
    primary_metric: str
    flops: Optional[int] = None
    params: Optional[int] = None
    total_evaluation_time: Optional[float] = None
    success: bool = False


@dataclass
class InferenceSummary:
    flops: Optional[int] = None
    params: Optional[int] = None
    total_inference_time: Optional[float] = None
    success: bool = False



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

import logging
from typing import TypedDict

import torch
import torch.nn as nn
from fvcore.nn import FlopCountAnalysis

from .environment import get_device


def get_params_and_flops(model, sample_input: torch.Tensor):
    if not isinstance(model, nn.Module): # Only support nn.Module
        return 0, 0
    sample_input = sample_input.to(get_device(model))
    # From v0.0.9
    flops, params = _params_and_flops_fvcore(model, sample_input)

    return flops, params

def _params_and_flops_fvcore(model: nn.Module, sample_input: torch.Tensor):
    fvcore_logger = logging.getLogger('fvcore')
    fvcore_logger.setLevel(logging.CRITICAL)
    # According to https://detectron2.readthedocs.io/en/latest/modules/fvcore.html#fvcore.nn.FlopCountAnalysis,
    # fvcore's one flop is equivalent to one multiply-add operation. This is considered as a MAC operation.
    flops = FlopCountAnalysis(model, sample_input).total() * 2 # Multiply by 2 to get FLOPs
    params = sum(p.numel() for p in model.parameters())
    return flops, params



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

"""
Based on the publicly available MixNet-PyTorch repository.
https://github.com/romulus0914/MixNet-PyTorch/blob/master/mixnet.py
"""
from collections import OrderedDict
import math
from typing import Dict, List, Optional

from omegaconf import DictConfig
import torch
from torch import nn
from torch.nn import functional as F
from torchvision.ops.misc import SqueezeExcitation as SEBlock

from ...op.registry import ACTIVATION_REGISTRY
from ...op.custom import ConvLayer
from ...utils import BackboneOutput
from ..registry import USE_INTERMEDIATE_FEATURES_TASK_LIST

__all__ = ['mixnet']

SUPPORTING_TASK = ['classification', 'segmentation', 'detection', 'pose_estimation']


# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
# GPConv: Grouped Point-wise Convolution for MixDepthBlock
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
class GPConv(nn.Module):
    def __init__(self, in_planes, out_planes, num_groups):
        super(GPConv, self).__init__()
        self.num_groups = num_groups
        assert in_planes % self.num_groups == 0
        sub_in_dim = in_planes // self.num_groups
        sub_out_dim = out_planes // self.num_groups

        self.group_point_wise = nn.ModuleList()
        for _ in range(self.num_groups):
            self.group_point_wise.append(nn.Conv2d(sub_in_dim, sub_out_dim,
                                                   kernel_size=1, stride=1, padding=0,
                                                   groups=1, dilation=1, bias=False))

    def forward(self, x):
        if self.num_groups == 1:
            return self.group_point_wise[0](x)

        chunks = torch.chunk(x, chunks=self.num_groups, dim=1)
        mix = [self.group_point_wise[stream](chunks[stream]) for stream in range(self.num_groups)]
        return torch.cat(mix, dim=1)


# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
# MDConv: Mixed Depth-wise Convolution for MixDepthBlock
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
class MDConv(nn.Module):
    def __init__(self, in_planes, kernel_sizes, stride=1, dilate=1):
        super(MDConv, self).__init__()
        self.num_groups = len(kernel_sizes)
        assert in_planes % self.num_groups == 0
        sub_hidden_dim = in_planes // self.num_groups

        assert stride in [1, 2]
        dilate = 1 if stride > 1 else dilate

        self.mixed_depth_wise = nn.ModuleList()
        for kernel_size in kernel_sizes:
            padding = ((kernel_size - 1) // 2) * dilate
            self.mixed_depth_wise.append(nn.Conv2d(sub_hidden_dim, sub_hidden_dim,
                                                   kernel_size=kernel_size, stride=stride, padding=padding,
                                                   groups=sub_hidden_dim, dilation=dilate, bias=False))

    def forward(self, x):
        if self.num_groups == 1:
            return self.mixed_depth_wise[0](x)

        chunks = torch.chunk(x, chunks=self.num_groups, dim=1)
        mix = [self.mixed_depth_wise[stream](chunks[stream]) for stream in range(self.num_groups)]
        return torch.cat(mix, dim=1)


# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
# MixDepthBlock: MixDepthBlock for MixNet
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
class MixDepthBlock(nn.Module):
    def __init__(self, in_planes, out_planes,
                 expand_ratio, num_exp_groups, kernel_sizes, num_poi_groups, stride, dilate,
                 reduction_ratio=4, dropout_rate=0.2, act_type="swish"):
        super(MixDepthBlock, self).__init__()
        self.dropout_rate = dropout_rate
        self.expand_ratio = expand_ratio
        self.out_channels = out_planes

        self.groups = len(kernel_sizes)
        self.use_se = (reduction_ratio is not None) and (reduction_ratio > 1)
        self.use_residual = in_planes == out_planes and stride == 1

        assert stride in [1, 2]
        dilate = 1 if stride > 1 else dilate
        hidden_dim = in_planes * expand_ratio

        # step 1. Expansion phase/Point-wise convolution
        if expand_ratio != 1:
            self.expansion = nn.Sequential(OrderedDict([
                ("conv", GPConv(in_planes, hidden_dim, num_groups=num_exp_groups)),
                ("norm", nn.BatchNorm2d(hidden_dim, eps=1e-3, momentum=0.01)),
                ("act", ACTIVATION_REGISTRY[act_type]())
            ]))

        # step 2. Depth-wise convolution phase
        self.depth_wise = nn.Sequential(OrderedDict([
            ("conv", MDConv(hidden_dim, kernel_sizes=kernel_sizes, stride=stride, dilate=dilate)),
            ("norm", nn.BatchNorm2d(hidden_dim, eps=1e-3, momentum=0.01)),
            ("act", ACTIVATION_REGISTRY[act_type]())
        ]))

        # step 3. Squeeze and Excitation
        if self.use_se:
            reduced_dim = max(1, int(in_planes / reduction_ratio))
            self.se_block = SEBlock(input_channels=hidden_dim, squeeze_channels=reduced_dim, activation=ACTIVATION_REGISTRY[act_type])

        # step 4. Point-wise convolution phase
        self.point_wise = nn.Sequential(OrderedDict([
            ("conv", GPConv(hidden_dim, out_planes, num_groups=num_poi_groups)),
            ("norm", nn.BatchNorm2d(out_planes, eps=1e-3, momentum=0.01))
        ]))

    def forward(self, x):
        res = x

        # step 1. Expansion phase/Point-wise convolution
        if self.expand_ratio != 1:
            x = self.expansion(x)

        # step 2. Depth-wise convolution phase
        x = self.depth_wise(x)

        # step 3. Squeeze and Excitation
        if self.use_se:
            x = self.se_block(x)

        # step 4. Point-wise convolution phase
        x = self.point_wise(x)

        # step 5. Skip connection and drop connect
        if self.use_residual:
            if self.training and (self.dropout_rate is not None):
                x = F.dropout2d(input=x, p=self.dropout_rate,
                                training=self.training, )
            x = x + res

        return x


class MixNet(nn.Module):
    def __init__(
        self,
        task: str,
        params: Optional[DictConfig] = None,
        stage_params: Optional[List] = None,
    ):
        # Check task compatibility
        self.task = task.lower()
        assert self.task in SUPPORTING_TASK, f'MixNet is not supported on {self.task} task now.'
        self.use_intermediate_features = self.task in USE_INTERMEDIATE_FEATURES_TASK_LIST

        super(MixNet, self).__init__()

        stem_channels = params.stem_channels
        width_multi = params.wid_mul
        depth_multi = params.dep_mul
        self.dropout_rate = params.dropout_rate

        out_channels = self._round_filters(stem_channels, width_multi)
        self.mod1 = ConvLayer(in_channels=3, out_channels=out_channels, kernel_size=3,
                              stride=2, groups=1, dilation=1, act_type="relu")

        in_channels = out_channels
        drop_rate = self.dropout_rate
        stages: List[nn.Module] = []
        for stg_idx, stage_info in enumerate(stage_params):
            
            stage: List[nn.Module] = []
            for block in zip(stage_info.expansion_ratio, stage_info.out_channels, stage_info.num_blocks,
                             stage_info.kernel_sizes, stage_info.num_exp_groups, stage_info.num_poi_groups,
                             stage_info.stride, stage_info.act_type, stage_info.se_reduction_ratio):
                t, c, n, k, ek, pk, s, a, se = block
                d = 1
                out_channels = self._round_filters(c, width_multi)
                repeats = self._round_repeats(n, depth_multi)

                for block_id in range(repeats):
                    stride = s if block_id == 0 else 1
                    dilate = d if stride == 1 else 1

                    stage.append(MixDepthBlock(in_channels, out_channels,
                                               expand_ratio=t, num_exp_groups=ek,
                                               kernel_sizes=k, num_poi_groups=pk,
                                               stride=stride, dilate=dilate,
                                               reduction_ratio=se,
                                               dropout_rate=drop_rate,
                                               act_type=a))

                    in_channels = out_channels
            
            # add last conv
            if stg_idx == len(stage_params) - 1:
                self.last_channels = 1536
                stage.append(
                    ConvLayer(in_channels=in_channels,
                              out_channels=self.last_channels,
                              kernel_size=1,
                              stride=1,
                              groups=1,
                              dilation=1,
                              act_type="relu")
                )
            
            stage = nn.Sequential(*stage)
            stages.append(stage)

        self.stages = nn.ModuleList(stages)
        self.avgpool = nn.AdaptiveAvgPool2d(1)

        self._feature_dim = self.last_channels
        self._intermediate_features_dim = [s[-1].out_channels for s in self.stages[:-1]]
        self._intermediate_features_dim += [self.last_channels]

        self._initialize_weights()

    def _initialize_weights(self):
        # weight initialization
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode='fan_out')
                if m.bias is not None:
                    nn.init.zeros_(m.bias)
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.ones_(m.weight)
                nn.init.zeros_(m.bias)
            elif isinstance(m, nn.Linear):
                fan_out = m.weight.size(0)
                init_range = 1.0 / math.sqrt(fan_out)
                nn.init.uniform_(m.weight, -init_range, init_range)
                if m.bias is not None:
                    nn.init.zeros_(m.bias)

    @staticmethod
    def _make_divisible(value, divisor=8):
        new_value = max(divisor, int(value + divisor / 2) // divisor * divisor)
        if new_value < 0.9 * value:
            new_value += divisor
        return new_value

    def _round_filters(self, filters, width_multi):
        if width_multi == 1.0:
            return filters
        return int(self._make_divisible(filters * width_multi))

    @staticmethod
    def _round_repeats(repeats, depth_multi):
        if depth_multi == 1.0:
            return repeats
        return int(math.ceil(depth_multi * repeats))
    
    @property
    def feature_dim(self):
        return self._feature_dim

    @property
    def intermediate_features_dim(self):
        return self._intermediate_features_dim

    def forward(self, x):
        x = self.mod1(x)

        all_hidden_states = () if self.use_intermediate_features else None
        for stage in self.stages:
            x = stage(x)
            if self.use_intermediate_features:
                all_hidden_states = all_hidden_states + (x, )
        
        if self.use_intermediate_features:
            return BackboneOutput(intermediate_features=all_hidden_states)

        x = self.avgpool(x)
        x = torch.flatten(x, 1)

        if self.training and (self.dropout_rate is not None):
            x = F.dropout(input=x, p=self.dropout_rate,
                          training=self.training, )

        return BackboneOutput(last_feature=x)


def mixnet(task, conf_model_backbone) -> MixNet:
    return MixNet(task, conf_model_backbone.params, conf_model_backbone.stage_params)


# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

from typing import Dict, Optional

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor
from torchvision.models.detection._utils import BoxCoder, Matcher
from torchvision.ops import boxes as box_ops

from ..common import SigmoidFocalLoss


class RetinaNetLoss(nn.Module):
    def __init__(self, **kwargs) -> None:
        super().__init__()
        fg_iou_thresh = 0.5
        bg_iou_thresh = 0.4
        self.proposal_matcher = Matcher(
                fg_iou_thresh,
                bg_iou_thresh,
                allow_low_quality_matches=True,
        )
        self.cls_loss = RetinaNetClassificationLoss(self.proposal_matcher.BETWEEN_THRESHOLDS)
        self.reg_loss = RetinaNetRegressionLoss()

    def compute_loss(self, out, target, matched_idxs):
        cls_loss = self.cls_loss(out, target, matched_idxs)
        box_loss = self.reg_loss(out, target, matched_idxs)

        # TODO: return as dict
        return cls_loss + box_loss

    def forward(self, out: Dict, target: Dict) -> torch.Tensor:
        matched_idxs = []
        anchors = out['anchors']

        for targets_per_image in target['gt']:
            if targets_per_image["boxes"].numel() == 0:
                matched_idxs.append(
                    torch.full((anchors.size(0),), -1, dtype=torch.int64, device=anchors.device)
                )
                continue

            match_quality_matrix = box_ops.box_iou(targets_per_image["boxes"], anchors)
            matched_idxs.append(self.proposal_matcher(match_quality_matrix))

        return self.compute_loss(out, target, matched_idxs)


class RetinaNetClassificationLoss(nn.Module):
    def __init__(self, between_threshold) -> None:
        super().__init__()
        self.between_threshold = between_threshold
        # TODO: Get from config
        alpha = 0.25
        gamma = 2
        self.focal_loss = SigmoidFocalLoss(alpha=alpha, gamma=gamma)

    def forward(self, out, target, matched_idxs):
        losses = []

        cls_logits = out["cls_logits"]
        cls_logits = torch.cat(cls_logits, dim=1)

        for targets_per_image, cls_logits_per_image, matched_idxs_per_image in zip(target['gt'], cls_logits, matched_idxs):
            # determine only the foreground
            foreground_idxs_per_image = matched_idxs_per_image >= 0
            num_foreground = foreground_idxs_per_image.sum()

            # create the target classification
            gt_classes_target = torch.zeros_like(cls_logits_per_image)
            gt_classes_target[
                foreground_idxs_per_image,
                targets_per_image["labels"][matched_idxs_per_image[foreground_idxs_per_image]],
            ] = 1.0

            # find indices for which anchors should be ignored
            valid_idxs_per_image = matched_idxs_per_image != self.between_threshold

            # compute the classification loss
            losses.append(
                self.focal_loss(
                    {'pred': cls_logits_per_image[valid_idxs_per_image]},
                    {'target': gt_classes_target[valid_idxs_per_image]},
                    reduction="sum",
                )
                / max(1, num_foreground)
            )

        return sum(losses) / len(target['gt'])


class RetinaNetRegressionLoss(nn.Module):
    def __init__(self) -> None:
        super().__init__()
        self.box_coder = BoxCoder(weights=(1.0, 1.0, 1.0, 1.0))

    def forward(self, out, target, matched_idxs):
        losses = []

        bbox_regression = out["bbox_regression"]
        bbox_regression = torch.cat(bbox_regression, dim=1)
        anchors = out["anchors"]

        for targets_per_image, bbox_regression_per_image, matched_idxs_per_image in zip(
            target['gt'], bbox_regression, matched_idxs
        ):
            # determine only the foreground indices, ignore the rest
            foreground_idxs_per_image = torch.where(matched_idxs_per_image >= 0)[0]
            num_foreground = foreground_idxs_per_image.numel()

            # select only the foreground boxes
            matched_gt_boxes_per_image = targets_per_image["boxes"][matched_idxs_per_image[foreground_idxs_per_image]]
            bbox_regression_per_image = bbox_regression_per_image[foreground_idxs_per_image, :]
            anchors_per_image = anchors[foreground_idxs_per_image, :]

            target_regression = self.box_coder.encode_single(matched_gt_boxes_per_image, anchors_per_image)
            loss = F.l1_loss(bbox_regression_per_image, target_regression, reduction="sum")

            # compute the loss
            losses.append(loss / max(1, num_foreground))

        return sum(losses) / max(1, len(target['gt']))



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

# Modified from FPN of mmdetection:
# https://github.com/open-mmlab/mmdetection/blob/main/mmdet/models/necks/fpn.py

from typing import List

from omegaconf import DictConfig
import torch 
import torch.nn as nn
import torch.nn.functional as F

from ...utils import BackboneOutput
from ...op.custom import ConvLayer


class FPN(nn.Module):

    def __init__(
        self,
        intermediate_features_dim: List[int],
        params: DictConfig,
    ):
        upsample_cfg = {'mode': 'nearest'}
        self.upsample_cfg = upsample_cfg.copy()
        super(FPN, self).__init__()

        self.in_channels = intermediate_features_dim
        self.out_channels = intermediate_features_dim[-1]
        self.num_ins = len(self.in_channels)
        self.num_outs = params.num_outs
        self.relu_before_extra_convs = params.relu_before_extra_convs
        self.add_extra_convs = params.add_extra_convs

        start_level = params.start_level
        end_level = params.end_level

        if end_level == -1 or end_level == self.num_ins - 1:
            self.backbone_end_level = self.num_ins
            assert self.num_outs >= self.num_ins - start_level
        else:
            # if end_level is not the last level, no extra level is allowed
            self.backbone_end_level = end_level + 1
            assert end_level < self.num_ins
            assert self.num_outs == end_level - start_level + 1
        self.start_level = start_level
        self.end_level = end_level
        self.add_extra_convs = self.add_extra_convs
        assert isinstance(self.add_extra_convs, (str, bool))
        if isinstance(self.add_extra_convs, str):
            # Extra_convs_source choices: 'on_input', 'on_lateral', 'on_output'
            assert self.add_extra_convs in ('on_input', 'on_lateral', 'on_output')
        elif self.add_extra_convs:  # True
            self.add_extra_convs = 'on_input'

        self.lateral_convs = nn.ModuleList()
        self.fpn_convs = nn.ModuleList()

        for i in range(self.start_level, self.backbone_end_level):
            l_conv = nn.Conv2d(self.in_channels[i], self.out_channels, kernel_size=1, stride=1, padding=0)
            # ConvModule(
            #     in_channels[i],
            #     out_channels,
            #     1,
            #     conv_cfg=conv_cfg,
            #     norm_cfg=norm_cfg if not self.no_norm_on_lateral else None,
            #     act_cfg=act_cfg,
            #     inplace=False)
            fpn_conv = nn.Conv2d(self.out_channels, self.out_channels, kernel_size=3, stride=1, padding=1)
            # ConvModule(
            #     out_channels,
            #     out_channels,
            #     3,
            #     padding=1,
            #     conv_cfg=conv_cfg,
            #     norm_cfg=norm_cfg,
            #     act_cfg=act_cfg,
            #     inplace=False)

            self.lateral_convs.append(l_conv)
            self.fpn_convs.append(fpn_conv)

        # add extra conv layers (e.g., RetinaNet)
        extra_levels = self.num_outs - self.backbone_end_level + self.start_level
        if self.add_extra_convs and extra_levels >= 1:
            for i in range(extra_levels):
                if i == 0 and self.add_extra_convs == 'on_input':
                    in_channels = self.in_channels[self.backbone_end_level - 1]
                else:
                    in_channels = self.out_channels
                extra_fpn_conv = nn.Conv2d(in_channels, self.out_channels, kernel_size=3, stride=2, padding=1)
                self.fpn_convs.append(extra_fpn_conv)

        self._intermediate_features_dim = [self.out_channels for _ in range(self.num_outs)]

    def forward(self, inputs):
        """Forward function."""
        assert len(inputs) == len(self.in_channels)

        # build laterals
        laterals = [
            lateral_conv(inputs[i + self.start_level])
            for i, lateral_conv in enumerate(self.lateral_convs)
        ]

        # build top-down path
        used_backbone_levels = len(laterals)
        for i in range(used_backbone_levels - 1, 0, -1):
            # In some cases, fixing `scale factor` (e.g. 2) is preferred, but
            #  it cannot co-exist with `size` in `F.interpolate`.
            if 'scale_factor' in self.upsample_cfg:
                # fix runtime error of "+=" inplace operation in PyTorch 1.10
                laterals[i - 1] = laterals[i - 1] + F.interpolate(
                    laterals[i], **self.upsample_cfg)
            else:
                prev_shape = laterals[i - 1].shape[2:]
                laterals[i - 1] = laterals[i - 1] + F.interpolate(
                    laterals[i], size=prev_shape, **self.upsample_cfg)

        # build outputs
        # part 1: from original levels
        outs = [
            self.fpn_convs[i](laterals[i]) for i in range(used_backbone_levels)
        ]
        # part 2: add extra levels
        if self.num_outs > len(outs):
            # use max pool to get more levels on top of outputs
            # (e.g., Faster R-CNN, Mask R-CNN)
            if not self.add_extra_convs:
                for i in range(self.num_outs - used_backbone_levels):
                    outs.append(F.max_pool2d(outs[-1], 1, stride=2))
            # add conv layers on top of original feature maps (RetinaNet)
            else:
                if self.add_extra_convs == 'on_input':
                    extra_source = inputs[self.backbone_end_level - 1]
                elif self.add_extra_convs == 'on_lateral':
                    extra_source = laterals[-1]
                elif self.add_extra_convs == 'on_output':
                    extra_source = outs[-1]
                else:
                    raise NotImplementedError
                outs.append(self.fpn_convs[used_backbone_levels](extra_source))
                for i in range(used_backbone_levels + 1, self.num_outs):
                    if self.relu_before_extra_convs:
                        outs.append(self.fpn_convs[i](F.relu(outs[-1])))
                    else:
                        outs.append(self.fpn_convs[i](outs[-1]))
        return BackboneOutput(intermediate_features=outs)
    
    @property
    def intermediate_features_dim(self):
        return self._intermediate_features_dim


def fpn(intermediate_features_dim, conf_model_neck, **kwargs):
    return FPN(intermediate_features_dim=intermediate_features_dim, params=conf_model_neck.params)


class LightFPN(nn.Module):
    def __init__(
        self,
        intermediate_features_dim: List[int], 
        params: DictConfig,
    ) -> None:
        super().__init__()

        self.in_channels = intermediate_features_dim 
        self.out_channels = params.out_channels
        self.num_ins = len(self.in_channels)
        self._intermediate_features_dim = [self.out_channels for _ in range(self.num_ins)]
        # TODO: Make sure this module can process multi-scale features greater than 2. 
        self.conv_C2 = ConvLayer(self.in_channels[0]+self.in_channels[1], self.out_channels, 1, 1, padding=0)
        self.conv_C3 = ConvLayer(self.in_channels[1], self.out_channels, 1, 1, padding=0)   
    
    def forward(self, inputs): 
        C2, C3 = inputs[0], inputs[1] 
        S3 = self.conv_C3(C3)
        P2 = F.interpolate(C3, scale_factor=2)
        P2 = torch.cat((P2, C2), dim=1)
        S2 = self.conv_C2(P2) 

        outs = [S2, S3] 

        return BackboneOutput(intermediate_features=outs)

    @property
    def intermediate_features_dim(self):
        return self._intermediate_features_dim

def lightfpn(intermediate_features_dim, conf_model_neck, **kwargs): 
    return LightFPN(intermediate_features_dim=intermediate_features_dim, params=conf_model_neck.params)


# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

import argparse
import json
import os
import subprocess
from dataclasses import dataclass, field
from pathlib import Path
from typing import List, Literal, Optional, Tuple, Union

import torch
from omegaconf import DictConfig, OmegaConf

from ..models import SUPPORTING_TASK_LIST
from ..models.utils import get_model_format

OUTPUT_ROOT_DIR = "./outputs"
LOG_LEVEL = os.getenv('LOG_LEVEL', 'INFO')


@dataclass
class ConfigSummary:
    task: Optional[str] = None
    model_name: Optional[str] = None
    is_graphmodule_training: Optional[bool] = None
    logging_dir: Optional[Path] = None


def str2bool(v):
    return v.lower() in ("yes", "true", "t", "1")


def parse_args_netspresso(with_gpus=False, isTrain=True):

    parser = argparse.ArgumentParser(description="Parser for NetsPresso configuration")

    # -------- User arguments ----------------------------------------

    if with_gpus:
        parser.add_argument(
            '--gpus', type=parse_gpu_ids, default="",
            dest='gpus',
            help='GPU device indices (comma-separated)')

    parser.add_argument(
        '--data', type=str, required=True,
        dest='data',
        help="Config for dataset information")

    parser.add_argument(
        '--augmentation', type=str, required=True,
        dest='augmentation',
        help="Config for data augmentation")

    parser.add_argument(
        '--model', type=str, required=True,
        dest='model',
        help="Config for the model architecture")

    if isTrain:
        parser.add_argument(
            '--training', type=str, required=True,
            dest='training',
            help="Config for training options")

    parser.add_argument(
        '--logging', type=str, default='config/logging.yaml',
        dest='logging',
        help="Config for logging options")

    parser.add_argument(
        '--environment', type=str, default='config/environment.yaml',
        dest='environment',
        help="Config for training environment (# workers, etc.)")

    parser.add_argument(
        '--log-level', choices=['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'], default=LOG_LEVEL,
        dest='log_level',
        help="Logging level in training process")

    parser.add_argument(
        '--task', type=str, default=None,
        dest='task',
        help="")

    parser.add_argument(
        '--model-name', type=str, default=None,
        dest='model_name',
        help="")

    parser.add_argument(
        '--is-graphmodule-training', type=str2bool, default=None,
        dest='is_graphmodule_training',
        help="")

    parser.add_argument(
        '--logging-dir', type=Path, default=None,
        dest='logging_dir',
        help="")

    args_parsed, _ = parser.parse_known_args()

    return args_parsed


def parse_gpu_ids(gpu_arg: str) -> Optional[Union[List, int]]:
    """Parse comma-separated GPU IDs and return as a list of integers."""
    if gpu_arg is None or str(gpu_arg) in ["", "None"]:
        return None

    try:
        gpu_ids = [int(id) for id in gpu_arg.split(',')]

        if len(gpu_ids) == 1:  # Single GPU
            return gpu_ids[0]

        gpu_ids = sorted(gpu_ids)
        return gpu_ids
    except ValueError as e:
        raise argparse.ArgumentTypeError('Invalid GPU IDs. Please provide comma-separated integers.') from e


def set_arguments(
    data: Union[Path, str],
    augmentation: Union[Path, str],
    model: Union[Path, str],
    logging: Union[Path, str],
    environment: Union[Path, str],
    training: Optional[Union[Path, str]] = None,
) -> DictConfig:

    conf_data = OmegaConf.load(data)
    conf_augmentation = OmegaConf.load(augmentation)
    conf_model = OmegaConf.load(model)
    if training:
        conf_training = OmegaConf.load(training)
    conf_logging = OmegaConf.load(logging)
    conf_environment = OmegaConf.load(environment)

    conf = OmegaConf.create()
    conf.merge_with(conf_data)
    conf.merge_with(conf_augmentation)
    conf.merge_with(conf_model)
    if training:
        conf.merge_with(conf_training)
    conf.merge_with(conf_logging)
    conf.merge_with(conf_environment)

    return conf


def get_gpu_from_config(conf_environment: DictConfig) -> Optional[Union[List, int]]:
    conf_environment_gpus = str(conf_environment.gpus) if hasattr(conf_environment, 'gpus') else None
    return parse_gpu_ids(conf_environment_gpus)


def get_gpus_from_parser_and_config(
    gpus: Optional[Union[List, int]],
    conf_environment: DictConfig
) -> Union[List, int]:
    conf_environment_gpus = get_gpu_from_config(conf_environment)
    if gpus is None:
        if conf_environment_gpus is None:
            return 0  # Try use the 'cuda:0'
        return conf_environment_gpus
    return gpus


def get_new_logging_dir(output_root_dir, project_id, mode: Literal['training', 'evaluation', 'inference']):
    version_idx = 0
    project_dir: Path = Path(output_root_dir) / project_id

    while (project_dir / f"version_{version_idx}").exists():
        version_idx += 1

    new_logging_dir: Path = project_dir / f"version_{version_idx}"
    new_logging_dir.mkdir(exist_ok=True, parents=True)

    summary_path = new_logging_dir / f"{mode}_summary.json"
    with open(summary_path, 'w') as f:
        json.dump({"status": "", "error_stats": ""}, f, indent=4)

    return new_logging_dir


def validate_train_config(conf: DictConfig) -> ConfigSummary:
    # Get information from configuration
    model_format = get_model_format(conf.model)
    assert model_format in ['torch', 'torch.fx'], "Training model format must be either 'torch' or 'torch.fx'. (.safetensors or .pt)"
    is_graphmodule_training = model_format == 'torch.fx'

    task = str(conf.model.task).lower()
    assert task in SUPPORTING_TASK_LIST

    model_name = str(conf.model.name).lower()

    if is_graphmodule_training:
        model_name += "_graphmodule"

    project_id = conf.logging.project_id if conf.logging.project_id is not None else f"{task}_{model_name}"
    logging_dir: Path = get_new_logging_dir(output_root_dir=conf.logging.output_dir, project_id=project_id, mode='training')

    return ConfigSummary(task=task, model_name=model_name, is_graphmodule_training=is_graphmodule_training, logging_dir=logging_dir)


def validate_evaluation_config(conf: DictConfig, gpus: Union[List, int]) -> ConfigSummary:

    task = str(conf.model.task).lower()
    assert task in SUPPORTING_TASK_LIST

    model_format = get_model_format(conf.model)
    if model_format == 'onnx':
        assert isinstance(gpus, int), "ONNX model evaluation must be done on a single GPU."

    model_name = str(conf.model.name).lower() + '_evaluation'

    project_id = conf.logging.project_id if conf.logging.project_id is not None else f"{task}_{model_name}"
    logging_dir: Path = get_new_logging_dir(output_root_dir=conf.logging.output_dir, project_id=project_id, mode='evaluation')

    return ConfigSummary(task=task, model_name=model_name, is_graphmodule_training=None, logging_dir=logging_dir)


def validate_inference_config(conf: DictConfig, gpus: Union[List, int]) -> ConfigSummary:

    task = str(conf.model.task).lower()
    assert task in SUPPORTING_TASK_LIST

    model_format = get_model_format(conf.model)
    if model_format == 'onnx':
        assert isinstance(gpus, int), "ONNX model inference must be done on a single GPU."

    model_name = str(conf.model.name).lower() + '_inference'

    project_id = conf.logging.project_id if conf.logging.project_id is not None else f"{task}_{model_name}"
    logging_dir: Path = get_new_logging_dir(output_root_dir=conf.logging.output_dir, project_id=project_id, mode='inference')

    return ConfigSummary(task=task, model_name=model_name, is_graphmodule_training=None, logging_dir=logging_dir)



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

from typing import Dict, List, Type

from .base import BasePipeline
from .evaluation import EvaluationPipeline
from .inference import InferencePipeline
from .task_processors.base import BaseTaskProcessor
from .task_processors.classification import ClassificationProcessor
from .task_processors.detection import DetectionProcessor
from .task_processors.pose_estimation import PoseEstimationProcessor
from .task_processors.segmentation import SegmentationProcessor
from .train import TrainingPipeline

# TODO: Temporary defined. It should be integrated with `..models.registry.SUPPORTING_TASK_LIST`
SUPPORTING_TASK_LIST: List[str] = ['classification', 'segmentation', 'detection', 'pose_estimation']

TASK_PROCESSOR: Dict[str, Type[BaseTaskProcessor]] = {
    'classification': ClassificationProcessor,
    'segmentation': SegmentationProcessor,
    'detection': DetectionProcessor,
    'pose_estimation': PoseEstimationProcessor,
}

PIPELINES: Dict[str, Type[BasePipeline]] = {
    'train': TrainingPipeline,
    'evaluation': EvaluationPipeline,
    'inference': InferencePipeline,
}



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

import math
from typing import List, Tuple, Union

from omegaconf import DictConfig
import torch
import torch.nn as nn
import torch.nn.functional as F

from ....op.custom import ConvLayer
from ....utils import FXTensorListType, ModelOutput


class AllMLPDecoder(nn.Module):
    def __init__(
            self,
            num_classes: int,
            intermediate_features_dim: List[int],
            params: DictConfig,
        ):
        super().__init__()
        intermediate_channels = params.intermediate_channels
        classifier_dropout_prob = params.classifier_dropout_prob

        # linear layers which will unify the channel dimension of each of the encoder blocks to the same config.decoder_hidden_size
        mlps = []
        for feature_dim in intermediate_features_dim:
            mlp = ConvLayer(feature_dim, intermediate_channels, kernel_size=1,
                            use_norm=True, use_act=True, norm_type='batch_norm', act_type='relu')
            mlps.append(mlp)
        self.linear_c = nn.ModuleList(mlps)

        # the following 3 layers implement the ConvModule of the original implementation
        self.linear_fuse = ConvLayer(len(intermediate_features_dim) * intermediate_channels,
                                     intermediate_channels, kernel_size=1,
                                     use_norm=True, use_act=True)

        self.dropout = nn.Dropout(classifier_dropout_prob)
        self.classifier = nn.Conv2d(intermediate_channels, num_classes, kernel_size=1)

    def forward(self, encoder_hidden_states: FXTensorListType, targets=None):

        all_hidden_states = ()
        for encoder_hidden_state, mlp in zip(encoder_hidden_states, self.linear_c):

            # unify channel dimension
            encoder_hidden_state = mlp(encoder_hidden_state)

            encoder_hidden_state = F.interpolate(
                encoder_hidden_state, size=encoder_hidden_states[0].size()[2:], mode="bilinear", align_corners=False
            )  # upsample to H/4 x W/4
            all_hidden_states += (encoder_hidden_state,)

        hidden_states = self.linear_fuse(torch.cat(all_hidden_states, dim=1))
        hidden_states = self.dropout(hidden_states)

        # logits are of shape (batch_size, num_labels, height/4, width/4)
        logits = self.classifier(hidden_states)  # B x {num_classes} x H/4 x W/4

        return ModelOutput(pred=logits)


def all_mlp_decoder(num_classes, intermediate_features_dim, conf_model_head, **kwargs) -> AllMLPDecoder:
    return AllMLPDecoder(num_classes,
                         intermediate_features_dim,
                         params=conf_model_head.params)



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

"""
This code is a modified version of https://github.com/roboflow/supervision/blob/a7edf467172df921608f0360112ba70e2259077c/supervision/metrics/detection.py
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Callable, List, Optional, Tuple

import numpy as np

from ..base import BaseMetric


def box_iou_batch(boxes_true: np.ndarray, boxes_detection: np.ndarray) -> np.ndarray:
    """
    Compute Intersection over Union (IoU) of two sets of bounding boxes -
        `boxes_true` and `boxes_detection`. Both sets
        of boxes are expected to be in `(x_min, y_min, x_max, y_max)` format.

    Args:
        boxes_true (np.ndarray): 2D `np.ndarray` representing ground-truth boxes.
            `shape = (N, 4)` where `N` is number of true objects.
        boxes_detection (np.ndarray): 2D `np.ndarray` representing detection boxes.
            `shape = (M, 4)` where `M` is number of detected objects.

    Returns:
        np.ndarray: Pairwise IoU of boxes from `boxes_true` and `boxes_detection`.
            `shape = (N, M)` where `N` is number of true objects and
            `M` is number of detected objects.
    """

    def box_area(box):
        return (box[2] - box[0]) * (box[3] - box[1])

    area_true = box_area(boxes_true.T)
    area_detection = box_area(boxes_detection.T)

    top_left = np.maximum(boxes_true[:, None, :2], boxes_detection[:, :2])
    bottom_right = np.minimum(boxes_true[:, None, 2:], boxes_detection[:, 2:])

    area_inter = np.prod(np.clip(bottom_right - top_left, a_min=0, a_max=None), 2)
    return area_inter / (area_true[:, None] + area_detection - area_inter)


def match_detection_batch(
    predictions: np.ndarray, targets: np.ndarray, iou_thresholds: np.ndarray
) -> np.ndarray:
    """
    Match predictions with target labels based on IoU levels.

    Args:
        predictions (np.ndarray): Batch prediction. Describes a single image and
            has `shape = (M, 6)` where `M` is the number of detected objects.
            Each row is expected to be in
            `(x_min, y_min, x_max, y_max, class, conf)` format.
        targets (np.ndarray): Batch target labels. Describes a single image and
            has `shape = (N, 5)` where `N` is the number of ground-truth objects.
            Each row is expected to be in
            `(x_min, y_min, x_max, y_max, class)` format.
        iou_thresholds (np.ndarray): Array contains different IoU thresholds.

    Returns:
        np.ndarray: Matched prediction with target labels result.
    """
    num_predictions, num_iou_levels = predictions.shape[0], iou_thresholds.shape[0]
    correct = np.zeros((num_predictions, num_iou_levels), dtype=bool)
    iou = box_iou_batch(targets[:, :4], predictions[:, :4])
    correct_class = targets[:, 4:5] == predictions[:, 4]

    for i, iou_level in enumerate(iou_thresholds):
        matched_indices = np.where((iou >= iou_level) & correct_class)

        if matched_indices[0].shape[0]:
            combined_indices = np.stack(matched_indices, axis=1)
            iou_values = iou[matched_indices][:, None]
            matches = np.hstack([combined_indices, iou_values])

            if matched_indices[0].shape[0] > 1:
                matches = matches[matches[:, 2].argsort()[::-1]]
                matches = matches[np.unique(matches[:, 1], return_index=True)[1]]
                matches = matches[np.unique(matches[:, 0], return_index=True)[1]]

            correct[matches[:, 1].astype(int), i] = True

    return correct


def compute_average_precision(recall: np.ndarray, precision: np.ndarray) -> float:
    """
    Compute the average precision using 101-point interpolation (COCO), given
        the recall and precision curves.

    Args:
        recall (np.ndarray): The recall curve.
        precision (np.ndarray): The precision curve.

    Returns:
        float: Average precision.
    """
    extended_recall = np.concatenate(([0.0], recall, [1.0]))
    extended_precision = np.concatenate(([1.0], precision, [0.0]))
    max_accumulated_precision = np.flip(
        np.maximum.accumulate(np.flip(extended_precision))
    )
    interpolated_recall_levels = np.linspace(0, 1, 101)
    interpolated_precision = np.interp(
        interpolated_recall_levels, extended_recall, max_accumulated_precision
    )
    average_precision = np.trapz(interpolated_precision, interpolated_recall_levels)
    return average_precision


def average_precisions_per_class(
    matches: np.ndarray,
    prediction_confidence: np.ndarray,
    prediction_class_ids: np.ndarray,
    true_class_ids: np.ndarray,
    eps: float = 1e-16,
) -> np.ndarray:
    """
    Compute the average precision, given the recall and precision curves.
    Source: https://github.com/rafaelpadilla/Object-Detection-Metrics.

    Args:
        matches (np.ndarray): True positives.
        prediction_confidence (np.ndarray): Objectness value from 0-1.
        prediction_class_ids (np.ndarray): Predicted object classes.
        true_class_ids (np.ndarray): True object classes.
        eps (float, optional): Small value to prevent division by zero.

    Returns:
        np.ndarray: Average precision for different IoU levels.
    """
    sorted_indices = np.argsort(-prediction_confidence)
    matches = matches[sorted_indices]
    prediction_class_ids = prediction_class_ids[sorted_indices]

    unique_classes, class_counts = np.unique(true_class_ids, return_counts=True)
    num_classes = unique_classes.shape[0]

    average_precisions = np.zeros((num_classes, matches.shape[1]))

    for class_idx, class_id in enumerate(unique_classes):
        is_class = prediction_class_ids == class_id
        total_true = class_counts[class_idx]
        total_prediction = is_class.sum()

        if total_prediction == 0 or total_true == 0:
            continue

        false_positives = (1 - matches[is_class]).cumsum(0)
        true_positives = matches[is_class].cumsum(0)
        recall = true_positives / (total_true + eps)
        precision = true_positives / (true_positives + false_positives)

        for iou_level_idx in range(matches.shape[1]):
            average_precisions[
                class_idx, iou_level_idx
            ] = compute_average_precision(
                recall[:, iou_level_idx], precision[:, iou_level_idx]
            )

    return average_precisions


class DetectionMetric(BaseMetric):
    SUPPORT_METRICS: List[str] = ['map50', 'map75', 'map50_95']

    def __init__(self, **kwargs):
        # TODO: Select metrics by user
        metric_names: List[str] = ['map50', 'map75', 'map50_95']
        primary_metric: str = 'map50'
        assert set(metric_names).issubset(DetectionMetric.SUPPORT_METRICS)
        super().__init__(metric_names=metric_names, primary_metric=primary_metric)

    def calibrate(self, predictions, targets, **kwargs):
        result_dict = {k: 0. for k in self.metric_names}

        iou_thresholds = np.linspace(0.5, 0.95, 10)
        stats = []

        # Gather matching stats for predictions and targets
        for pred, target in zip(predictions, targets):
            predicted_objs_bbox, predicted_objs_class, predicted_objs_confidence = pred['post_boxes'], pred['post_labels'], pred['post_scores']
            true_objs_bbox, true_objs_class = target['boxes'], target['labels']

            true_objs = np.concatenate((true_objs_bbox, true_objs_class[..., np.newaxis]), axis=-1)
            predicted_objs = np.concatenate((predicted_objs_bbox, predicted_objs_class[..., np.newaxis], predicted_objs_confidence[..., np.newaxis]), axis=-1)

            if predicted_objs.shape[0] == 0 and true_objs.shape[0]:
                stats.append(
                    (
                        np.zeros((0, iou_thresholds.size), dtype=bool),
                        *np.zeros((2, 0)),
                        true_objs[:, 4],
                    )
                )

            if true_objs.shape[0]:
                matches = match_detection_batch(predicted_objs, true_objs, iou_thresholds)
                stats.append(
                    (
                        matches,
                        predicted_objs[:, 5],
                        predicted_objs[:, 4],
                        true_objs[:, 4],
                    )
                )

        # Compute average precisions if any matches exist
        if stats:
            concatenated_stats = [np.concatenate(items, 0) for items in zip(*stats)]
            average_precisions = average_precisions_per_class(*concatenated_stats)
            #result_dict['map50'] = average_precisions[:, 0].mean()
            #result_dict['map75'] = average_precisions[:, 5].mean()
            #result_dict['map50_95'] = average_precisions.mean()
            self.metric_meter['map50'].update(average_precisions[:, 0].mean())
            self.metric_meter['map75'].update(average_precisions[:, 5].mean())
            self.metric_meter['map50_95'].update(average_precisions.mean())
        else:
            #result_dict['map50'], result_dict['map75'], result_dict['map50_95'] = 0, 0, 0
            self.metric_meter['map50'].update(0)
            self.metric_meter['map75'].update(0)
            self.metric_meter['map50_95'].update(0)

        return result_dict



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

import os
from abc import abstractmethod
from typing import Callable, Optional, Tuple, Union

import torch
import torch.nn as nn
from loguru import logger
from omegaconf import OmegaConf

from .utils import BackboneOutput, DetectionModelOutput, ModelOutput


class TaskModel(nn.Module):
    def __init__(self, conf_model, backbone, neck, head, freeze_backbone: bool = False) -> None:
        super(TaskModel, self).__init__()
        self.task = conf_model.task
        self.name = conf_model.name
        self.backbone_name = conf_model.architecture.backbone.name
        if neck:
            self.neck_name = conf_model.architecture.neck.name
        self.head_name = conf_model.architecture.head.name

        self.backbone = backbone
        if neck:
            self.neck = neck
        self.head = head

        if freeze_backbone:
            self._freeze_backbone()
            logger.info(f"Freeze! {self.backbone_name} is now freezed. Now only tuning with {self.head_name}.")

    def _freeze_backbone(self):
        for m in self.backbone.parameters():
            m.requires_grad = False

    @property
    def head_list(self):
        return ('head')

    @property
    def device(self):
        return next(self.parameters()).device

    def _get_name(self):
        if hasattr(self, 'neck'):
            return f"{self.__class__.__name__}[task={self.task}, backbone={self.backbone_name}, neck={self.neck_name}, head={self.head_name}]"
        else:
            return f"{self.__class__.__name__}[task={self.task}, backbone={self.backbone_name}, head={self.head_name}]"

    @abstractmethod
    def forward(self, x, label_size=None, targets=None):
        raise NotImplementedError

    def deploy(self, ):
        self.eval()
        for m in self.modules():
            if hasattr(m, 'convert_to_deploy'):
                m.convert_to_deploy()
        return self


class ClassificationModel(TaskModel):
    def __init__(self, conf_model, backbone, neck, head, freeze_backbone=False) -> None:
        super().__init__(conf_model, backbone, neck, head, freeze_backbone)

    def forward(self, x, label_size=None, targets=None):
        features: BackboneOutput = self.backbone(x)
        out: ModelOutput = self.head(features['last_feature'])
        return out


class SegmentationModel(TaskModel):
    def __init__(self, conf_model, backbone, neck, head, freeze_backbone=False) -> None:
        super().__init__(conf_model, backbone, neck, head, freeze_backbone)

    def forward(self, x, label_size=None, targets=None):
        features: BackboneOutput = self.backbone(x)
        if hasattr(self, 'neck'):
            features: BackboneOutput = self.neck(features['intermediate_features'])
        out: ModelOutput = self.head(features['intermediate_features'], targets)
        return out


class DetectionModel(TaskModel):
    def __init__(self, conf_model, backbone, neck, head, freeze_backbone=False) -> None:
        super().__init__(conf_model, backbone, neck, head, freeze_backbone)

    def forward(self, x, label_size=None, targets=None):
        features: BackboneOutput = self.backbone(x)
        if hasattr(self, 'neck'):
            features: BackboneOutput = self.neck(features['intermediate_features'])
        out: DetectionModelOutput = self.head(features['intermediate_features'], targets)
        return out


class PoseEstimationModel(TaskModel):
    def __init__(self, conf_model, backbone, neck, head, freeze_backbone=False) -> None:
        super().__init__(conf_model, backbone, neck, head, freeze_backbone)

    def forward(self, x, label_size=None, targets=None):
        features: BackboneOutput = self.backbone(x)
        if hasattr(self, 'neck'):
            features: BackboneOutput = self.neck(features['intermediate_features'])
        out: DetectionModelOutput = self.head(features['intermediate_features'], targets)
        return out


class ONNXModel:
    '''
        ONNX Model wrapper class for inferencing.
    '''
    def __init__(self, model_conf) -> None:
        import onnxruntime as ort
        self.name = model_conf.name + '_onnx'
        self.onnx_path = model_conf.checkpoint.path
        providers = [
            ('CUDAExecutionProvider', {
                'device_id': 0,
            }),
            'CPUExecutionProvider',
        ]
        self.inference_session = ort.InferenceSession(model_conf.checkpoint.path, providers=providers)

    def _get_name(self):
        return f"{self.__class__.__name__}[model={self.name}]"

    def __call__(self, x, label_size=None, targets=None):
        device = x.device
        x = x.detach().cpu().numpy()
        out = self.inference_session.run(None, {self.inference_session.get_inputs()[0].name: x})
        out = [torch.tensor(o).to(device) for o in out]

        if len(out) == 1:
            out = out[0]

        return ModelOutput(pred=out)

    def eval(self):
        pass # Do nothing

    def set_provider(self, device):
        if device.type == 'cuda':
            self.inference_session.set_providers(['CUDAExecutionProvider'])
        else:
            self.inference_session.set_providers(['CPUExecutionProvider'])



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

from omegaconf import OmegaConf

from .registry import SCHEDULER_DICT


def build_scheduler(optimizer, training_conf):
    scheduler_conf = training_conf.scheduler
    scheduler_name = scheduler_conf.name
    num_epochs = training_conf.epochs

    assert scheduler_name in SCHEDULER_DICT, f"{scheduler_name} not in scheduler dict!"
    lr_scheduler = SCHEDULER_DICT[scheduler_name](optimizer, scheduler_conf, num_epochs)

    return lr_scheduler, num_epochs



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

# from .core import *
from .core.resnet import resnet
from .experimental.darknet import cspdarknet
from .experimental.efficientformer import efficientformer
from .experimental.mixnet import mixnet
from .experimental.mixtransformer import mixtransformer
from .experimental.mobilenetv3 import mobilenetv3
from .experimental.mobilenetv4 import mobilenetv4
from .experimental.mobilevit import mobilevit
from .experimental.shufflenetv2 import shufflenetv2
from .experimental.vit import vit



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

import argparse
import math
import warnings
from functools import partial
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import torch
import torch.nn as nn
from torch import Tensor
from torch.fx.proxy import Proxy
from torchvision.ops.misc import SqueezeExcitation as SElayer

from ..op.registry import ACTIVATION_REGISTRY, NORM_REGISTRY


def make_divisible(
    v: Union[float, int],
    divisor: Optional[int] = 8,
    min_value: Optional[Union[float, int]] = None,
) -> Union[float, int]:
    """
    This function is taken from the original tf repo.
    It ensures that all layers have a channel number that is divisible by 8
    It can be seen here:
    https://github.com/tensorflow/models/blob/master/research/slim/nets/mobilenet/mobilenet.py
    :param v:
    :param divisor:
    :param min_value:
    :return:
    """
    if min_value is None:
        min_value = divisor
    new_v = max(min_value, int(v + divisor / 2) // divisor * divisor)
    # Make sure that round down does not go down by more than 10%.
    if new_v < 0.9 * v:
        new_v += divisor
    return new_v


class ConvLayer(nn.Module):

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: Union[int, Tuple[int, int]],
        stride: Optional[Union[int, Tuple[int, int]]] = 1,
        dilation: Optional[Union[int, Tuple[int, int]]] = 1,
        padding: Optional[Union[int, Tuple[int, int]]] = None,
        groups: Optional[int] = 1,
        bias: bool = False,
        padding_mode: Optional[str] = 'zeros',
        use_norm: bool = True,
        norm_type: Optional[str] = None,
        use_act: bool = True,
        act_type: Optional[str] = None,
    ) -> None:
        super().__init__()

        if isinstance(kernel_size, int):
            kernel_size = (kernel_size, kernel_size)

        if isinstance(stride, int):
            stride = (stride, stride)

        if isinstance(dilation, int):
            dilation = (dilation, dilation)

        assert isinstance(kernel_size, Tuple)
        assert isinstance(stride, Tuple)
        assert isinstance(dilation, Tuple)

        if padding is None:
            padding = (
                int((kernel_size[0] - 1) / 2) * dilation[0],
                int((kernel_size[1] - 1) / 2) * dilation[1],
            )

        block = nn.Sequential()

        conv_layer = nn.Conv2d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            groups=groups,
            bias=bias,
            padding_mode=padding_mode,
        )

        block.add_module(name='conv', module=conv_layer)

        self.norm_name = None
        if use_norm or norm_type is not None:
            _norm_type = norm_type.lower() if norm_type is not None else 'batch_norm'
            if bias:
                warnings.warn("Bias would be ignored in batch normalization", stacklevel=2)
            assert _norm_type in NORM_REGISTRY
            cls_norm = NORM_REGISTRY[_norm_type]
            norm_layer = cls_norm(num_features=out_channels)
            block.add_module(name='norm', module=norm_layer)
            self.norm_name = norm_layer.__class__.__name__

        self.act_name = None
        if use_act or act_type is not None:
            _act_type = act_type.lower() if act_type is not None else 'relu'
            assert _act_type in ACTIVATION_REGISTRY
            cls_act = ACTIVATION_REGISTRY[_act_type]
            act_layer = cls_act()
            block.add_module(name='act', module=act_layer)
            self.act_name = act_layer.__class__.__name__

        self.block = block

    def forward(self, x: Union[Tensor, Proxy]) -> Union[Tensor, Proxy]:
        return self.block(x)

    def __repr__(self):
        return f"{self.block}"


class SeparableConvLayer(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: Union[int, Tuple[int, int]],
        stride: Optional[Union[int, Tuple[int, int]]] = 1,
        dilation: Optional[Union[int, Tuple[int, int]]] = 1,
        padding: Optional[Union[int, Tuple[int, int]]] = None,
        bias: bool = False,
        padding_mode: Optional[str] = 'zeros',
        use_norm: bool = True,
        norm_type: Optional[str] = None,
        use_act: bool = True,
        act_type: Optional[str] = None,
        no_out_act: Optional[bool] = False,
    ) -> None:
        super().__init__()
        if act_type is None:
            act_type = 'relu'
        self.depthwise = ConvLayer(in_channels=in_channels, out_channels=in_channels,
                                   kernel_size=kernel_size, stride=stride, dilation=dilation,
                                   padding=padding, groups=in_channels, bias=bias, padding_mode=padding_mode,
                                   use_norm=use_norm, norm_type=norm_type, use_act=use_act, act_type=act_type,)
        self.pointwise = ConvLayer(in_channels=in_channels, out_channels=out_channels, kernel_size=1,
                                   use_norm=use_norm, norm_type=norm_type, use_act=False)
        self.final_act = nn.Identity() if no_out_act else ACTIVATION_REGISTRY[act_type]()

    def forward(self, x: Union[Tensor, Proxy]) -> Union[Tensor, Proxy]:
        x = self.depthwise(x)
        x = self.pointwise(x)
        x = self.final_act(x)
        return x


class BasicBlock(nn.Module):
    expansion: int = 1

    def __init__(
        self,
        inplanes: int,
        planes: int,
        stride: int = 1,
        downsample: Optional[nn.Module] = None,
        groups: int = 1,
        base_width: int = 64,
        dilation: int = 1,
        norm_layer: Optional[str] = None,
        expansion: Optional[int] = None,
        act_type: Optional[str] = None,
        no_out_act: bool = False
    ) -> None:
        super(BasicBlock, self).__init__()
        if norm_layer is None:
            norm_layer = 'batch_norm'
        if act_type is None:
            act_type = 'relu'
        if groups != 1 or base_width != 64:
            raise ValueError('BasicBlock only supports groups=1 and base_width=64')
        if dilation > 1:
            raise NotImplementedError("Dilation > 1 not supported in BasicBlock")
        if expansion is not None:
            self.expansion = expansion
        # Both self.conv1 and self.downsample layers downsample the input when stride != 1

        self.conv1 = ConvLayer(in_channels=inplanes, out_channels=planes,
                               kernel_size=3, stride=stride, dilation=1, padding=1, groups=1,
                               norm_type=norm_layer, act_type=act_type)

        self.conv2 = ConvLayer(in_channels=planes, out_channels=planes,
                               kernel_size=3, stride=1, dilation=1, padding=1, groups=1,
                               norm_type=norm_layer, use_act=False)

        self.downsample = downsample
        self.final_act = nn.Identity() if no_out_act else ACTIVATION_REGISTRY[act_type]()

    def forward(self, x: Tensor) -> Tensor:
        identity = x

        out = self.conv1(x)
        out = self.conv2(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.final_act(out)

        return out


class Bottleneck(nn.Module):
    expansion: int = 4

    def __init__(
        self,
        inplanes: int,
        planes: int,
        stride: int = 1,
        downsample: Optional[nn.Module] = None,
        groups: int = 1,
        base_width: int = 64,
        dilation: int = 1,
        norm_layer: Optional[str] = None,
        expansion: Optional[int] = None,
        act_type: Optional[str] = None,
        no_out_act: bool = False
    ) -> None:
        super(Bottleneck, self).__init__()
        if norm_layer is None:
            norm_layer = 'batch_norm'
        if act_type is None:
            act_type = 'relu'
        width = int(planes * (base_width / 64.)) * groups
        if expansion is not None:
            self.expansion = expansion
        # Both self.conv2 and self.downsample layers downsample the input when stride != 1

        self.conv1 = ConvLayer(in_channels=inplanes, out_channels=width,
                               kernel_size=1, stride=1,
                               norm_type=norm_layer, act_type=act_type)

        self.conv2 = ConvLayer(in_channels=width, out_channels=width,
                               kernel_size=3, stride=stride, dilation=dilation, padding=dilation, groups=groups,
                               norm_type=norm_layer, act_type=act_type)

        self.conv3 = ConvLayer(in_channels=width, out_channels=planes * self.expansion,
                               kernel_size=1, stride=1,
                               norm_type=norm_layer, use_act=False)

        self.downsample = downsample
        self.final_act = nn.Identity() if no_out_act else ACTIVATION_REGISTRY[act_type]()

    def forward(self, x: Tensor) -> Tensor:
        identity = x

        out = self.conv1(x)
        out = self.conv2(out)
        out = self.conv3(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.final_act(out)

        return out


class InvertedResidual(nn.Module):
    # Implemented as described at section 5 of MobileNetV3 paper
    def __init__(
        self,
        in_channels: int,
        hidden_channels: int,
        out_channels: int,
        kernel_size: Union[int, Tuple[int, int]],
        stride: Optional[Union[int, Tuple[int, int]]] = 1,
        dilation: Optional[Union[int, Tuple[int, int]]] = 1,
        norm_type: Optional[str] = None,
        act_type: Optional[str] = None,
        use_se: bool = False,
        se_layer: Callable[..., nn.Module] = partial(SElayer, scale_activation=nn.Hardsigmoid),
    ):
        super().__init__()
        if not (1 <= stride <= 2):
            raise ValueError("illegal stride value")

        self.use_res_connect = stride == 1 and in_channels == out_channels

        layers: List[nn.Module] = []

        # expand
        if hidden_channels != in_channels:
            layers.append(
                ConvLayer(
                    in_channels=in_channels,
                    out_channels=hidden_channels,
                    kernel_size=1,
                    norm_type=norm_type,
                    act_type=act_type,
                )
            )

        # depthwise
        stride = 1 if dilation > 1 else stride
        layers.append(
            ConvLayer(
                in_channels=hidden_channels,
                out_channels=hidden_channels,
                kernel_size=kernel_size,
                stride=stride,
                dilation=dilation,
                groups=hidden_channels,
                norm_type=norm_type,
                act_type=act_type,
            )
        )
        if use_se:
            squeeze_channels = make_divisible(hidden_channels // 4, 8)
            layers.append(se_layer(hidden_channels, squeeze_channels))

        # project
        layers.append(
            ConvLayer(
                in_channels=hidden_channels,
                out_channels=out_channels,
                kernel_size=1,
                norm_type=norm_type,
                use_act=False
            )
        )

        self.block = nn.Sequential(*layers)
        self.out_channels = out_channels
        self._is_cn = stride > 1

    def forward(self, input: Tensor) -> Tensor:
        result = self.block(input)
        if self.use_res_connect:
            result = result + input
        return result


class UniversalInvertedResidualBlock(nn.Module):
    # Based on MobileNetV4: https://arxiv.org/pdf/2404.10518
    def __init__(
        self,
        in_channels: int,
        hidden_channels: int,
        out_channels: int,
        extra_dw: bool = False,
        extra_dw_kernel_size: Optional[Union[int, Tuple[int, int]]] = None,
        middle_dw: bool = True,
        middle_dw_kernel_size: Union[int, Tuple[int, int]] = 3,
        stride: Optional[Union[int, Tuple[int, int]]] = 1,
        norm_type: Optional[str] = None,
        act_type: Optional[str] = None,
        use_se: bool = False,
        se_layer: Callable[..., nn.Module] = partial(SElayer, scale_activation=nn.Hardsigmoid),
        layer_scale: Optional[float] = None,
    ):
        super().__init__()
        if not (1 <= stride <= 2):
            raise ValueError("illegal stride value")

        self.use_res_connect = stride == 1 and in_channels == out_channels
        layers: List[nn.Module] = []

        # extra depthwise conv
        if extra_dw:
            assert extra_dw_kernel_size is not None, "if extra_dw is True, extra_dw_kernel_size must be provided."
            layers.append(
                ConvLayer(
                    in_channels=in_channels,
                    out_channels=in_channels,
                    kernel_size=extra_dw_kernel_size,
                    stride=1,
                    groups=in_channels,
                    norm_type=norm_type,
                    use_act=False # No activation for extra depthwise conv
                )
            )

        # expand
        if hidden_channels != in_channels:
            layers.append(
                ConvLayer(
                    in_channels=in_channels,
                    out_channels=hidden_channels,
                    kernel_size=1,
                    norm_type=norm_type,
                    act_type=act_type,
                )
            )

        # middle depthwise
        if middle_dw:
            assert middle_dw_kernel_size is not None, "if middle_dw is True, middle_dw_kernel_size must be provided."
            layers.append(
                ConvLayer(
                    in_channels=hidden_channels,
                    out_channels=hidden_channels,
                    kernel_size=middle_dw_kernel_size,
                    stride=stride,
                    groups=hidden_channels,
                    norm_type=norm_type,
                    act_type=act_type,
                )
            )

        if use_se:
            squeeze_channels = make_divisible(hidden_channels // 4, 8)
            layers.append(se_layer(hidden_channels, squeeze_channels))

        # project
        layers.append(
            ConvLayer(
                in_channels=hidden_channels,
                out_channels=out_channels,
                kernel_size=1,
                norm_type=norm_type,
                use_act=False # No activation for project conv
            )
        )

        self.block = nn.Sequential(*layers)
        self.apply_layer_scale = False
        if layer_scale is not None:
            self.apply_layer_scale = True
            self.layer_scale = LayerScale2d(out_channels, layer_scale)
        self.out_channels = out_channels
        self._is_cn = stride > 1

    def forward(self, input: Tensor) -> Tensor:
        result = self.block(input)
        if self.apply_layer_scale:
            result = self.layer_scale(result)
        if self.use_res_connect:
            result = result + input
        return result


class SinusoidalPositionalEncoding(nn.Module):
    """
    This layer adds sinusoidal positional embeddings to a 3D input tensor. The code has been adapted from
    `Pytorch tutorial <https://pytorch.org/tutorials/beginner/transformer_tutorial.html>`_

    Args:
        d_model (int): dimension of the input tensor
        dropout (Optional[float]): Dropout rate. Default: 0.0
        max_len (Optional[int]): Max. number of patches (or seq. length). Default: 5000
        channels_last (Optional[bool]): Channels dimension is the last in the input tensor

    Shape:
        - Input: :math:`(N, C, P)` or :math:`(N, P, C)` where :math:`N` is the batch size, :math:`C` is the embedding dimension,
            :math:`P` is the number of patches
        - Output: same shape as the input

    """

    def __init__(
        self,
        d_model: int,
        max_len: Optional[int] = 5000,
        channels_last: Optional[bool] = True,
        *args,
        **kwargs
    ) -> None:

        position_last = not channels_last

        pos_encoding = torch.zeros(max_len, d_model)
        position = torch.arange(0, max_len, dtype=torch.float).unsqueeze(1)
        div_term = torch.exp(
            torch.arange(0, d_model, 2).float() * (-math.log(10000.0) / d_model)
        )
        pos_encoding[:, 0::2] = torch.sin(position * div_term)
        pos_encoding[:, 1::2] = torch.cos(position * div_term)
        # add dummy batch dimension
        pos_encoding = pos_encoding.unsqueeze(0)  # [1 x C x P_max)

        patch_dim = -2  # patch dimension is second last (N, P, C)
        if position_last:
            pos_encoding = pos_encoding.transpose(
                1, 2
            )  # patch dimension is last (N, C, P)
            patch_dim = -1  # patch dimension is last (N, C, P)

        super().__init__()

        self.patch_dim = patch_dim
        self.register_buffer("pe", pos_encoding)

    def forward_patch_last(
        self, x, indices: Optional[Tensor] = None, *args, **kwargs
    ) -> Tensor:
        # seq_length should be the last dim
        if indices is None:
            x = x + self.pe[..., : x.shape[-1]]
        else:
            ndim = x.ndim
            repeat_size = [x.shape[0]] + [-1] * (ndim - 1)

            pe = self.pe.expand(repeat_size)
            selected_pe = torch.gather(pe, index=indices, dim=-1)
            x = x + selected_pe
        return x

    def forward_others(
        self, x, indices: Optional[Tensor] = None, *args, **kwargs
    ) -> Tensor:
        # seq_length should be the second last dim

        # @deepkyu: [fx tracing] Always `indices` is None
        # if indices is None:
        #     x = x + self.pe[..., : x.shape[-2], :]
        # else:
        #     ndim = x.ndim
        #     repeat_size = [x.shape[0]] + [-1] * (ndim - 1)

        #     pe = self.pe.expand(repeat_size)
        #     selected_pe = torch.gather(pe, index=indices, dim=-2)
        #     x = x + selected_pe

        # x = x + self.pe[..., :seq_index, :]
        if not isinstance(x, torch.fx.Proxy):
            self.last_token_num = x.shape[-2]
        x = x + self.pe[..., : self.last_token_num, :]

        return x

    def forward(self, x, indices: Optional[Tensor] = None, *args, **kwargs) -> Tensor:
        if self.patch_dim == -1:
            return self.forward_patch_last(x, indices=indices)
        else:
            return self.forward_others(x, indices=indices)

    # def profile_module(self, input: Tensor) -> Tuple[Tensor, float, float]:
    #     return input, 0.0, 0.0

    # def __repr__(self):
    #     return "{}(dropout={})".format(self.__class__.__name__, self.dropout.p)


class GlobalPool(nn.Module):
    """
    This layers applies global pooling over a 4D or 5D input tensor

    Args:
        pool_type (Optional[str]): Pooling type. It can be mean, rms, or abs. Default: `mean`
        keep_dim (Optional[bool]): Do not squeeze the dimensions of a tensor. Default: `False`

    Shape:
        - Input: :math:`(N, C, H, W)` or :math:`(N, C, D, H, W)`
        - Output: :math:`(N, C, 1, 1)` or :math:`(N, C, 1, 1, 1)` if keep_dim else :math:`(N, C)`
    """

    pool_types = ["mean", "rms", "abs"]

    def __init__(
        self,
        pool_type: Optional[str] = "mean",
        keep_dim: Optional[bool] = False,
        *args,
        **kwargs
    ) -> None:
        super().__init__()
        # if pool_type not in self.pool_types:
        #     logger.error(
        #         "Supported pool types are: {}. Got {}".format(
        #             self.pool_types, pool_type
        #         )
        #     )
        self.pool_type = pool_type
        self.keep_dim = keep_dim

    @classmethod
    def add_arguments(cls, parser: argparse.ArgumentParser):
        cls_name = "{} arguments".format(cls.__name__)
        group = parser.add_argument_group(title=cls_name, description=cls_name)
        group.add_argument(
            "--model.layer.global-pool",
            type=str,
            default="mean",
            help="Which global pooling?",
        )
        return parser

    def _global_pool(self, x: Tensor, dims: List):
        if self.pool_type == "rms":  # root mean square
            x = x**2
            x = torch.mean(x, dim=dims, keepdim=self.keep_dim)
            x = x**-0.5
        elif self.pool_type == "abs":  # absolute
            x = torch.mean(torch.abs(x), dim=dims, keepdim=self.keep_dim)
        else:
            # default is mean
            # same as AdaptiveAvgPool
            x = torch.mean(x, dim=dims, keepdim=self.keep_dim)
        return x

    def forward(self, x: Tensor) -> Tensor:
        # @deepkyu: [fx tracing] Always x.dim() == 4
        # if x.dim() == 4:
        #     dims = [-2, -1]
        # elif x.dim() == 5:
        #     dims = [-3, -2, -1]
        # else:
        #     raise NotImplementedError("Currently 2D and 3D global pooling supported")

        return self._global_pool(x, dims=(-2, -1))

    # def profile_module(self, input: Tensor) -> Tuple[Tensor, float, float]:
    #     input = self.forward(input)
    #     return input, 0.0, 0.0

    # def __repr__(self):
    #     return "{}(type={})".format(self.__class__.__name__, self.pool_type)


class Focus(nn.Module):
    """Focus width and height information into channel space."""

    def __init__(self, in_channels, out_channels, ksize=1, stride=1, act_type="silu"):
        super().__init__()
        self.conv = ConvLayer(in_channels=in_channels * 4,
                              out_channels=out_channels,
                              kernel_size=ksize,
                              stride=stride,
                              act_type=act_type)

    def forward(self, x):
        # shape of x (b,c,w,h) -> y(b,4c,w/2,h/2)
        patch_top_left = x[..., ::2, ::2]
        patch_top_right = x[..., ::2, 1::2]
        patch_bot_left = x[..., 1::2, ::2]
        patch_bot_right = x[..., 1::2, 1::2]
        x = torch.cat(
            (
                patch_top_left,
                patch_bot_left,
                patch_top_right,
                patch_bot_right,
            ),
            dim=1,
        )
        return self.conv(x)


class CSPLayer(nn.Module):
    """C3 in yolov5, CSP Bottleneck with 3 convolutions"""

    def __init__(
        self,
        in_channels,
        out_channels,
        n=1,
        shortcut=True,
        expansion=0.5,
        depthwise=False,
        act_type="silu",
    ):
        """
        Args:
            in_channels (int): input channels.
            out_channels (int): output channels.
            n (int): number of Bottlenecks. Default value: 1.
        """
        # ch_in, ch_out, number, shortcut, groups, expansion
        super().__init__()
        hidden_channels = int(out_channels * expansion)  # hidden channels
        self.conv1 = ConvLayer(in_channels=in_channels,
                               out_channels=hidden_channels,
                               kernel_size=1,
                               stride=1, act_type=act_type)
        self.conv2 = ConvLayer(in_channels=in_channels,
                              out_channels=hidden_channels,
                              kernel_size=1,
                              stride=1, act_type=act_type)
        self.conv3 = ConvLayer(in_channels=2 * hidden_channels,
                               out_channels=out_channels,
                               kernel_size=1,
                               stride=1, act_type=act_type)

        block = DarknetBlock

        module_list = [
            block(
                in_channels=hidden_channels,
                out_channels=hidden_channels,
                shortcut=shortcut,
                expansion=1.0,
                depthwise=depthwise,
                act_type=act_type
            )
            for _ in range(n)
        ]
        self.m = nn.Sequential(*module_list)

    def forward(self, x):
        x_1 = self.conv1(x)
        x_2 = self.conv2(x)
        x_1 = self.m(x_1)
        x = torch.cat((x_1, x_2), dim=1)
        return self.conv3(x)


class SPPBottleneck(nn.Module):
    """Spatial pyramid pooling layer used in YOLOv3-SPP"""

    def __init__(
        self, in_channels, out_channels, kernel_sizes=(5, 9, 13), act_type="silu"
    ):
        super().__init__()
        hidden_channels = in_channels // 2
        self.conv1 = ConvLayer(in_channels=in_channels, out_channels=hidden_channels,
                               kernel_size=1, stride=1, act_type=act_type)
        self.m = nn.ModuleList(
            [
                nn.MaxPool2d(kernel_size=ks, stride=1, padding=ks // 2)
                for ks in kernel_sizes
            ]
        )
        conv2_channels = hidden_channels * (len(kernel_sizes) + 1)
        self.conv2 = ConvLayer(in_channels=conv2_channels, out_channels=out_channels,
                               kernel_size=1, stride=1, act_type=act_type)

    def forward(self, x):
        x = self.conv1(x)
        x = torch.cat([x] + [m(x) for m in self.m], dim=1)
        x = self.conv2(x)
        return x


class ShuffleV2Block(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        hidden_channels: int,
        kernel_size: int,
        stride: int,
    ):
        super().__init__()
        assert stride in [1, 2], "Stride must be either 1 or 2"

        self.stride = stride
        self.in_channels = in_channels
        self.hidden_channels = hidden_channels
        self.kernel_size = kernel_size
        self.padding = kernel_size // 2

        self.out_channels = out_channels - in_channels

        self.branch_main = self._create_main_branch()
        self.branch_proj = self._create_proj_branch() if stride == 2 else None

    def _create_main_branch(self) -> nn.Sequential:
        return nn.Sequential(
            ConvLayer(self.in_channels, self.hidden_channels, 1, 1, padding=0),
            ConvLayer(self.hidden_channels, self.hidden_channels, self.kernel_size,
                      stride=self.stride, padding=self.padding, groups=self.hidden_channels, use_act=False),
            ConvLayer(self.hidden_channels, self.out_channels, 1, 1, padding=0),
        )

    def _create_proj_branch(self) -> nn.Sequential:
        return nn.Sequential(
            ConvLayer(self.in_channels, self.in_channels, self.kernel_size,
                      self.stride, padding=self.padding, groups=self.in_channels, use_act=False),
            ConvLayer(self.in_channels, self.in_channels, 1, 1, padding=0),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if self.stride == 1:
            x_proj, x = self.channel_shuffle(x)
            return torch.cat((x_proj, self.branch_main(x)), 1)
        else:
            return torch.cat((self.branch_proj(x), self.branch_main(x)), 1)

    def channel_shuffle(self, x: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        b, c, h, w = x.shape
        x = x.reshape(b * c // 2, 2, h * w)
        x = x.permute(1, 0, 2)
        x = x.reshape(2, -1, c // 2, h, w)
        return x[0], x[1]


# Newly defined because of slight difference with Bottleneck of custom.py
class DarknetBlock(nn.Module):
    # Standard bottleneck
    def __init__(
        self,
        in_channels,
        out_channels,
        shortcut=True,
        expansion=0.5,
        depthwise=False,
        act_type="silu",
    ):
        super().__init__()
        hidden_channels = int(out_channels * expansion)
        self.conv1 = ConvLayer(in_channels=in_channels, out_channels=hidden_channels,
                                kernel_size=1, stride=1, act_type=act_type)
        if depthwise:
            self.conv2 = SeparableConvLayer(in_channels=hidden_channels, out_channels=out_channels,
                                            kernel_size=3, stride=1, act_type=act_type)
        else:
            self.conv2 = ConvLayer(in_channels=hidden_channels, out_channels=out_channels,
                                    kernel_size=3, stride=1, act_type=act_type)
        self.use_add = shortcut and in_channels == out_channels

    def forward(self, x):
        y = self.conv2(self.conv1(x))
        if self.use_add:
            y = y + x
        return y


class LayerScale2d(nn.Module):
    """
        Based on timm implementation.
    """
    def __init__(self, dim: int, init_values: float = 1e-5, inplace: bool = False):
        super().__init__()
        self.inplace = inplace
        self.gamma = nn.Parameter(init_values * torch.ones(dim))

    def forward(self, x):
        gamma = self.gamma.view(1, -1, 1, 1)
        return x.mul_(gamma) if self.inplace else x * gamma



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

from typing import List

import numpy as np

from ..base import BaseMetric

IGNORE_INDEX_NONE_VALUE = 255


#TODO: Temporarily added IoUMeter. MetricMeter should be more generalized.
class IoUMeter(object):
    """Computes and stores the average and current value"""

    def __init__(self, num_classes, name: str, fmt=':f'):
        self.name = name
        self.fmt = fmt
        self.num_classes = num_classes
        self.reset()

    def reset(self):
        self.intersection = np.zeros(self.num_classes)
        self.union = np.zeros(self.num_classes)

    def update(self, intersection, union) -> None:
        self.intersection += intersection
        self.union += union

    def __str__(self):
        return f'{self.name} {np.nanmean(self.intersection / self.union):6.2f}'

    @property
    def avg(self) -> float:
        return np.nanmean(self.intersection / self.union)


class SegmentationMetric(BaseMetric):
    SUPPORT_METRICS: List[str] = ['iou', 'pixel_acc']

    def __init__(self, num_classes=None, ignore_index=IGNORE_INDEX_NONE_VALUE):
        # TODO: Select metrics by user
        metric_names = ['iou', 'pixel_acc']
        primary_metric = 'iou'

        assert set(metric_names).issubset(SegmentationMetric.SUPPORT_METRICS)
        super().__init__(metric_names=metric_names, primary_metric=primary_metric)
        self.ignore_index = ignore_index if ignore_index is not None else IGNORE_INDEX_NONE_VALUE
        self.K = num_classes

        self.metric_meter['iou'] = IoUMeter(num_classes=self.K, name='iou')

    def intersection_and_union(self, output, target):

        # 'K' classes, output and target sizes are N or N * L or N * H * W, each value in range 0 to K - 1.
        assert (len(output.shape) in [1, 2, 3])

        assert output.shape == target.shape
        output = output.reshape(-1)
        target = target.reshape(-1)
        output[target == self.ignore_index] = self.ignore_index
        intersection = output[output == target]
        #area_intersection = torch.histc(intersection, bins=self.K, min=0, max=self.K-1)
        #area_output = torch.histc(output, bins=self.K, min=0, max=self.K-1)
        #area_target = torch.histc(target, bins=self.K, min=0, max=self.K-1)
        area_intersection = np.histogram(intersection, bins=np.linspace(0, self.K, self.K+1))[0]
        area_output = np.histogram(output, bins=np.linspace(0, self.K, self.K+1))[0]
        area_target = np.histogram(target, bins=np.linspace(0, self.K, self.K+1))[0]
        area_union = area_output + area_target - area_intersection

        intersection, union, target, output = area_intersection, area_union, area_target, area_output

        return {
            'intersection': intersection,
            'union': union,
            'target': target,
            'output': output
        }

    def calibrate(self, pred, target, **kwargs):
        B = pred.shape[0]

        metrics = self.intersection_and_union(pred, target)
        self.metric_meter['iou'].update(metrics['intersection'], metrics['union'])
        self.metric_meter['pixel_acc'].update(sum(metrics['intersection']) / (sum(metrics['target']) + 1e-10), n=B)



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

from abc import ABC, abstractmethod
from typing import Dict, List, Literal, Optional, final

import torch
import torch.distributed as dist
from loguru import logger

NUM_SAMPLES = 16


class BaseTaskProcessor(ABC):
    def __init__(self, conf, postprocessor, devices, **kwargs):
        super(BaseTaskProcessor, self).__init__()
        self.conf = conf
        self.postprocessor = postprocessor
        self.devices = devices
        self.single_gpu_or_rank_zero = (not conf.distributed) or (conf.distributed and dist.get_rank() == 0)

        #TODO: Temporarily set ``mixed_precision`` as optional since this is experimental
        if hasattr(conf, 'training'):
            self.mixed_precision = conf.training.mixed_precision if hasattr(conf.training, 'mixed_precision') else False
            self.max_norm = conf.training.max_norm if hasattr(conf.training, 'max_norm') else None
        else:
            self.mixed_precision = False
            self.max_norm = None
        if self.mixed_precision:
            if self.single_gpu_or_rank_zero:
                logger.info("-" * 40)
                logger.info("Mixed precision training activated.")
            self.data_type = torch.float16
        else:
            self.data_type = torch.float32
        self.grad_scaler = torch.cuda.amp.GradScaler(enabled=self.mixed_precision)

    @abstractmethod
    def train_step(self, train_model, batch):
        raise NotImplementedError

    @abstractmethod
    def valid_step(self, eval_model, batch):
        raise NotImplementedError

    @abstractmethod
    def test_step(self, test_model, batch):
        raise NotImplementedError

    @abstractmethod
    def get_metric_with_all_outputs(self, outputs, phase: Literal['train', 'valid'], metric_factory):
        raise NotImplementedError



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

from .common import CrossEntropyLoss, SigmoidFocalLoss
from .detection import DETRLoss, RetinaNetLoss, YOLOXLoss
from .pose_estimation import RTMCCLoss
from .segmentation import PIDNetLoss, SegCrossEntropyLoss

LOSS_DICT = {
    'cross_entropy': CrossEntropyLoss,
    'seg_cross_entropy': SegCrossEntropyLoss,
    'pidnet_loss': PIDNetLoss,
    'yolox_loss': YOLOXLoss,
    'retinanet_loss': RetinaNetLoss,
    'detr_loss': DETRLoss,
    'focal_loss': SigmoidFocalLoss,
    'rtmcc_loss': RTMCCLoss,
}

PHASE_LIST = ['train', 'valid', 'test']
